# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import json
import datetime
from typer.testing import CliRunner
from unittest.mock import MagicMock
from rich.table import Table


from dorsal.cli import app
from dorsal.common.exceptions import DorsalClientError

runner = CliRunner()


mock_collection_record = MagicMock()
mock_collection_record.collection_id = "col_abc123"
mock_collection_record.name = "My Research Data"
mock_collection_record.file_count = 152
mock_collection_record.total_size = 545259520  # 520 MiB
mock_collection_record.is_private = True
mock_collection_record.date_modified = datetime.datetime(2025, 8, 9, 10, 30)

mock_pagination = MagicMock()
mock_pagination.page_count = 1
mock_pagination.current_page = 1
mock_pagination.start_index = 1
mock_pagination.end_index = 1
mock_pagination.record_count = 1
mock_pagination.has_next = False

MOCK_API_RESPONSE = MagicMock()
MOCK_API_RESPONSE.records = [mock_collection_record]
MOCK_API_RESPONSE.pagination = mock_pagination
MOCK_API_RESPONSE.model_dump_json.return_value = json.dumps(
    {
        "records": [{"collection_id": "col_abc123", "name": "My Research Data"}],
        "pagination": {"page_count": 1, "record_count": 1},
    }
)


@pytest.fixture
def mock_list_collections_cmd(mocker):
    """Mocks dependencies for the `collection list` command."""
    # Patch list_collections at its source due to lazy loading in the command
    mock_list = mocker.patch("dorsal.api.collection.list_collections", return_value=MOCK_API_RESPONSE)
    return mock_list


def test_list_collections_success_table_output(mock_rich_console, mock_list_collections_cmd):
    """Tests the default command, expecting a Rich Table."""
    result = runner.invoke(app, ["collection", "list"])

    assert result.exit_code == 0
    mock_list_collections_cmd.assert_called_once_with(page=1, per_page=25, mode="pydantic")

    # The command prints a status message, then the table
    assert mock_rich_console.print.call_count == 2
    printed_object = mock_rich_console.print.call_args.args[0]
    assert isinstance(printed_object, Table)
    assert "DorsalHub Collections" in str(printed_object.title)


def test_list_collections_with_pagination(mock_rich_console, mock_list_collections_cmd):
    """Tests that the pagination footer is displayed correctly."""
    mock_list_collections_cmd.return_value.pagination.page_count = 5
    mock_list_collections_cmd.return_value.pagination.has_next = True

    result = runner.invoke(app, ["collection", "list"])

    assert result.exit_code == 0
    # Combine all printed output to check for the footer
    all_printed_text = "".join(str(call.args[0]) for call in mock_rich_console.print.call_args_list)
    assert "Showing page 1 of 5" in all_printed_text
    assert "To see the next page, run the command again with --page 2" in all_printed_text


def test_list_collections_no_results(mock_rich_console, mock_list_collections_cmd):
    """Tests the output when no collections are found."""

    mock_list_collections_cmd.return_value.records = []

    result = runner.invoke(app, ["collection", "list"])

    assert result.exit_code == 0
    info_message = mock_rich_console.print.call_args.args[0]
    assert "No collections found" in info_message


def test_list_collections_json_output(mock_rich_console, mock_list_collections_cmd):
    """Tests the --json output."""
    result = runner.invoke(app, ["collection", "list", "--json"])

    assert result.exit_code == 0
    # Verifies the model's dump method was called and the result printed
    mock_list_collections_cmd.return_value.model_dump_json.assert_called_once()
    mock_rich_console.print.assert_called_once_with(mock_list_collections_cmd.return_value.model_dump_json.return_value)


def test_list_collections_api_error(mock_list_collections_cmd):
    """Tests graceful failure on a DorsalClientError."""
    mock_list_collections_cmd.side_effect = DorsalClientError("Authentication failed")

    result = runner.invoke(app, ["collection", "list"])

    assert result.exit_code != 0
    assert "API Error: Authentication failed" in result.output



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pathlib
from unittest.mock import MagicMock

import pytest
import typer
from typer.testing import CliRunner

from dorsal.cli.adapter_app.parse_cmd import parse_adapter
from dorsal.cli.themes.palettes import DEFAULT_PALETTE
from dorsal.common.exceptions import DorsalError

cli_app = typer.Typer()


@cli_app.callback()
def main_callback(ctx: typer.Context):
    ctx.obj = {"palette": DEFAULT_PALETTE}


cli_app.command(name="parse")(parse_adapter)

runner = CliRunner()


@pytest.fixture
def mock_parse_deps(mocker):
    """Mocks backend dependencies for the `adapter parse` command."""
    mock_rich_console = MagicMock()
    mock_error_console = MagicMock()

    mocker.patch("dorsal.common.cli.get_rich_console", return_value=mock_rich_console)
    mocker.patch("dorsal.common.cli.get_error_console", return_value=mock_error_console)

    mock_parse_api = mocker.patch("dorsal.api.adapters.parse_file_from_path")
    mock_parse_api.return_value = {
        "producer": "mock-adapter",
        "text": "Hello world",
        "segments": [],
    }

    return {
        "parse_api": mock_parse_api,
        "rich_console": mock_rich_console,
        "error_console": mock_error_console,
    }


def test_parse_cmd_success(mock_parse_deps):
    """Tests standard successful file parsing and saving with an explicit format."""
    with runner.isolated_filesystem():
        test_file = pathlib.Path("test.srt")
        test_file.write_text("1\n00:00:00,000 --> 00:00:01,000\nHello world", encoding="utf-8")

        result = runner.invoke(cli_app, ["parse", str(test_file), "open/audio-transcription", "--format", "srt"])

        assert result.exit_code == 0

        mock_parse_deps["parse_api"].assert_called_once()
        kwargs = mock_parse_deps["parse_api"].call_args.kwargs
        assert kwargs["schema_id"] == "open/audio-transcription"
        assert kwargs["source_format"] == "srt"

        printed_json = mock_parse_deps["rich_console"].print.call_args.args[0]
        data = json.loads(printed_json)
        assert data["schema_id"] == "open/audio-transcription"
        assert data["record"]["text"] == "Hello world"

        expected_save_path = pathlib.Path.cwd() / "test.dorsal.json"
        assert expected_save_path.exists()
        saved_data = json.loads(expected_save_path.read_text(encoding="utf-8"))
        assert saved_data["record"]["text"] == "Hello world"


def test_parse_cmd_infer_format(mock_parse_deps):
    """Tests that omitting the format flag correctly relies on the API's inference."""
    with runner.isolated_filesystem():
        test_file = pathlib.Path("test.srt")
        test_file.write_text("mock content", encoding="utf-8")

        result = runner.invoke(cli_app, ["parse", str(test_file), "open/audio-transcription"])

        assert result.exit_code == 0

        mock_parse_deps["parse_api"].assert_called_once()
        kwargs = mock_parse_deps["parse_api"].call_args.kwargs
        assert kwargs.get("source_format") is None


def test_parse_cmd_with_options(mock_parse_deps):
    """Tests that CLI options are correctly inferred and threaded to the parser."""
    with runner.isolated_filesystem():
        test_file = pathlib.Path("test.txt")
        test_file.write_text("Hello world", encoding="utf-8")

        result = runner.invoke(
            cli_app,
            [
                "parse",
                str(test_file),
                "open/document-extraction",
                "--format",
                "txt",
                "--opt",
                "strict=true",
                "-o",
                "chunk=5",
            ],
        )

        assert result.exit_code == 0

        kwargs = mock_parse_deps["parse_api"].call_args.kwargs
        assert kwargs["strict"] is True
        assert kwargs["chunk"] == 5


def test_parse_cmd_no_save_flag(mock_parse_deps):
    """Tests that --no-save prevents file creation."""
    with runner.isolated_filesystem():
        test_file = pathlib.Path("test.vtt")
        test_file.write_text("WEBVTT\n\nHello", encoding="utf-8")

        result = runner.invoke(
            cli_app, ["parse", str(test_file), "open/audio-transcription", "--format", "vtt", "--no-save"]
        )

        assert result.exit_code == 0

        expected_save_path = pathlib.Path.cwd() / "test.dorsal.json"
        assert not expected_save_path.exists()

        assert mock_parse_deps["rich_console"].print.called


def test_parse_cmd_missing_adapters(mock_parse_deps, mocker):
    """Tests handling of an environment missing the adapters package."""
    mock_parse_deps["parse_api"].side_effect = DorsalError("Please pip install dorsalhub-adapters to enable exports.")

    with runner.isolated_filesystem():
        test_file = pathlib.Path("test.srt")
        test_file.write_text("mock content", encoding="utf-8")

        result = runner.invoke(cli_app, ["parse", str(test_file), "open/audio-transcription", "--format", "srt"])

        assert result.exit_code != 0
        error_msg = str(mock_parse_deps["error_console"].print.call_args.args[0])
        assert "Parse Failed" in error_msg
        assert "dorsalhub-adapters" in error_msg


def test_parse_cmd_dorsal_error(mock_parse_deps):
    """Tests that specific parsing errors are caught and displayed nicely."""
    mock_parse_deps["parse_api"].side_effect = DorsalError("Failed to parse record from fake_format: unsupported")

    with runner.isolated_filesystem():
        test_file = pathlib.Path("test.fake")
        test_file.write_text("mock content", encoding="utf-8")

        result = runner.invoke(
            cli_app, ["parse", str(test_file), "open/audio-transcription", "--format", "fake_format"]
        )

        assert result.exit_code != 0
        error_msg = str(mock_parse_deps["error_console"].print.call_args.args[0])
        assert "unsupported" in error_msg



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pytest
import typer
from unittest.mock import MagicMock
from typer.testing import CliRunner

from dorsal.cli.annotation_app.get_annotation_cmd import get_annotation
from dorsal.cli.themes.palettes import DEFAULT_PALETTE
from dorsal.common.exceptions import DorsalError, NotFoundError, DorsalClientError

cli_app = typer.Typer()


@cli_app.callback()
def main_callback(ctx: typer.Context):
    ctx.obj = {"palette": DEFAULT_PALETTE}


cli_app.command(name="get")(get_annotation)
runner = CliRunner()


@pytest.fixture
def mock_get_deps(mocker, mock_rich_console):
    mock_error_console = MagicMock()
    mocker.patch("dorsal.common.cli.get_rich_console", return_value=mock_rich_console)
    mocker.patch("dorsal.common.cli.get_error_console", return_value=mock_error_console)

    mock_api_get = mocker.patch("dorsal.api.file.get_file_annotation")
    mock_panel = mocker.patch("dorsal.cli.views.model.create_model_result_panel")

    return {
        "api_get": mock_api_get,
        "panel": mock_panel,
        "error_console": mock_error_console,
    }


def test_get_annotation_success(mock_rich_console, mock_get_deps):
    mock_result = MagicMock()
    mock_result.schema_id = "AudioTranscription"
    mock_get_deps["api_get"].return_value = mock_result
    mock_get_deps["panel"].return_value = "PanelOutput"

    result = runner.invoke(cli_app, ["get", "uuid-1234"])

    assert result.exit_code == 0
    mock_get_deps["api_get"].assert_called_with("uuid-1234", mode="pydantic")
    mock_get_deps["panel"].assert_called_once_with(
        result=mock_result, title="AudioTranscription", file_name="ID: uuid-1234", palette=DEFAULT_PALETTE
    )
    assert mock_rich_console.print.called


def test_get_annotation_json(mock_rich_console, mock_get_deps):
    mock_get_deps["api_get"].return_value = '{"data": "raw"}'

    result = runner.invoke(cli_app, ["get", "uuid-1234", "--json"])

    assert result.exit_code == 0
    mock_get_deps["api_get"].assert_called_with("uuid-1234", mode="json")
    mock_rich_console.print.assert_called_with('{"data": "raw"}', end="\n")
    mock_get_deps["panel"].assert_not_called()


def test_get_annotation_error_panel(mock_get_deps):
    mock_get_deps["api_get"].side_effect = DorsalError("Annotation not found")

    result = runner.invoke(cli_app, ["get", "uuid-1234"])

    assert result.exit_code != 0
    error_msg = str(mock_get_deps["error_console"].print.call_args.args[0])
    assert "Annotation not found" in error_msg


def test_get_annotation_error_json(mock_get_deps):
    mock_get_deps["api_get"].side_effect = Exception("Internal crash")

    result = runner.invoke(cli_app, ["get", "uuid-1234", "--json"])

    assert result.exit_code != 0
    error_json = mock_get_deps["error_console"].print.call_args.args[0]
    data = json.loads(error_json)

    assert data["error"] == "Unexpected Error"
    assert data["detail"] == "Internal crash"


def test_get_annotation_exit_propagation(mock_get_deps):
    mock_get_deps["api_get"].side_effect = typer.Exit(0)
    result = runner.invoke(cli_app, ["get", "uuid-1234"])
    assert result.exit_code == 0


def test_get_annotation_not_found_panel(mock_get_deps):
    mock_get_deps["api_get"].side_effect = NotFoundError(message="Annotation 'uuid-1234' not found.")

    result = runner.invoke(cli_app, ["get", "uuid-1234"])

    assert result.exit_code != 0
    error_msg = str(mock_get_deps["error_console"].print.call_args.args[0])
    assert "Not Found:" in error_msg
    assert "Annotation 'uuid-1234' not found." in error_msg


def test_get_annotation_not_found_json(mock_get_deps):
    mock_get_deps["api_get"].side_effect = NotFoundError(message="Annotation 'uuid-1234' not found.")

    result = runner.invoke(cli_app, ["get", "uuid-1234", "--json"])

    assert result.exit_code != 0
    error_json = mock_get_deps["error_console"].print.call_args.args[0]
    data = json.loads(error_json)

    assert data["error"] == "Not Found"
    assert data["detail"] == "Annotation 'uuid-1234' not found."


def test_get_annotation_client_error_panel(mock_get_deps):
    mock_get_deps["api_get"].side_effect = DorsalClientError(message="Rate limit exceeded.")

    result = runner.invoke(cli_app, ["get", "uuid-1234"])

    assert result.exit_code != 0
    error_msg = str(mock_get_deps["error_console"].print.call_args.args[0])
    assert "API Error:" in error_msg
    assert "Rate limit exceeded." in error_msg


def test_get_annotation_client_error_json(mock_get_deps):
    mock_get_deps["api_get"].side_effect = DorsalClientError(message="Rate limit exceeded.")

    result = runner.invoke(cli_app, ["get", "uuid-1234", "--json"])

    assert result.exit_code != 0
    error_json = mock_get_deps["error_console"].print.call_args.args[0]
    data = json.loads(error_json)

    assert data["error"] == "API Error"
    assert data["detail"] == "Rate limit exceeded."


def test_get_annotation_mutually_exclusive_flags():
    """Covers the Typer BadParameter raise when both --json and --export are used."""
    result = runner.invoke(cli_app, ["get", "uuid-1234", "--json", "--export", "csl"])

    assert result.exit_code != 0
    assert "at the same time" in result.output


def test_get_annotation_export_success(mocker, mock_rich_console, mock_get_deps):
    """Covers successful export logic using the hydrated.record attribute."""
    mock_result = MagicMock()
    mock_result.schema_id = "Article"
    mock_result.record = {"title": "Test Paper"}
    mock_get_deps["api_get"].return_value = mock_result

    mock_export = mocker.patch("dorsal.api.adapters.export_record", return_value="exported_csl_string")

    result = runner.invoke(cli_app, ["get", "uuid-1234", "--export", "csl"])

    assert result.exit_code == 0
    mock_export.assert_called_once_with(record={"title": "Test Paper"}, schema_id="Article", target_format="csl")
    mock_rich_console.print.assert_called_with("exported_csl_string", end="")


def test_get_annotation_export_fallback_model_dump(mocker, mock_rich_console, mock_get_deps):
    """Covers the fallback logic when hydrated.record is missing but model_dump() exists."""
    mock_result = MagicMock()
    del mock_result.record
    mock_result.schema_id = "Article"
    mock_result.model_dump.return_value = {"record": {"title": "Fallback Data"}}
    mock_get_deps["api_get"].return_value = mock_result

    mock_export = mocker.patch("dorsal.api.adapters.export_record", return_value="exported_fallback_string")

    result = runner.invoke(cli_app, ["get", "uuid-1234", "--export", "csl"])

    assert result.exit_code == 0
    mock_export.assert_called_once_with(record={"title": "Fallback Data"}, schema_id="Article", target_format="csl")


def test_get_annotation_export_missing_record(mock_get_deps):
    """Covers the ValueError raise when no record data is found for export, and its JSON error format."""
    mock_result = MagicMock()
    mock_result.record = {}
    mock_result.model_dump.return_value = {}
    mock_get_deps["api_get"].return_value = mock_result

    result = runner.invoke(cli_app, ["get", "uuid-1234", "--export", "csl"])

    assert result.exit_code != 0
    error_json = mock_get_deps["error_console"].print.call_args.args[0]
    data = json.loads(error_json)
    assert data["error"] == "Data Error"
    assert "No record data found" in data["detail"]


def test_get_annotation_value_error_ui(mock_get_deps):
    """Covers the ValueError handler for standard UI output (no --json or --export)."""
    mock_get_deps["api_get"].side_effect = ValueError("Corrupt file")

    result = runner.invoke(cli_app, ["get", "uuid-1234"])

    assert result.exit_code != 0
    error_msg = str(mock_get_deps["error_console"].print.call_args.args[0])
    assert "Data Error:" in error_msg
    assert "Corrupt file" in error_msg


def test_get_annotation_output_file_standard(tmp_path, mock_get_deps, mock_rich_console):
    """Covers saving standard output directly to a file and dumping the pydantic model."""
    out_file = tmp_path / "custom_out.json"

    mock_result = MagicMock()
    mock_result.schema_id = "Article"
    mock_result.model_dump.return_value = {"id": "uuid-1234", "schema_id": "Article"}
    mock_get_deps["api_get"].return_value = mock_result
    mock_get_deps["panel"].return_value = "PanelOutput"

    result = runner.invoke(cli_app, ["get", "uuid-1234", "--output", str(out_file)])

    assert result.exit_code == 0
    assert out_file.exists()

    saved_data = json.loads(out_file.read_text(encoding="utf-8"))
    assert saved_data["id"] == "uuid-1234"
    assert mock_rich_console.print.called


def test_get_annotation_output_dir_json(tmp_path, mock_get_deps):
    """Covers saving pure JSON output to a directory (auto-resolving filename to .json)."""
    out_dir = tmp_path / "out_folder"
    out_dir.mkdir()

    mock_get_deps["api_get"].return_value = '{"raw": "json"}'

    result = runner.invoke(cli_app, ["get", "uuid-1234", "--json", "--output", str(out_dir)])

    assert result.exit_code == 0
    expected_file = out_dir / "uuid-1234.json"
    assert expected_file.exists()
    assert expected_file.read_text(encoding="utf-8") == '{"raw": "json"}'


def test_get_annotation_output_dir_export(mocker, tmp_path, mock_get_deps):
    """Covers saving exported output to a directory (auto-resolving extension)."""
    out_dir = tmp_path / "out_folder"
    out_dir.mkdir()

    mock_result = MagicMock()
    mock_result.schema_id = "Article"
    mock_result.record = {"title": "Test"}
    mock_get_deps["api_get"].return_value = mock_result

    mocker.patch("dorsal.api.adapters.export_record", return_value="exported text")
    mocker.patch("dorsal.api.adapters.get_format_extension", return_value="csl")

    result = runner.invoke(cli_app, ["get", "uuid-1234", "--export", "csl", "--output", str(out_dir)])

    assert result.exit_code == 0
    expected_file = out_dir / "uuid-1234.csl"
    assert expected_file.exists()
    assert expected_file.read_text(encoding="utf-8") == "exported text"



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import pathlib
from unittest.mock import patch
from dorsal.registry import initialize
from dorsal.common.exceptions import DorsalError


def test_create_project_success(tmp_path):
    """
    Test that a new project is created with the correct structure and content.
    """

    project_name = "My Cool Scanner"
    result = initialize.create_new_annotation_model_project(project_name, target_dir=tmp_path)

    expected_clean_name = "my-cool-scanner"
    expected_pkg_name = "dorsal-my-cool-scanner"

    assert result.clean_name == expected_clean_name
    assert result.package_name == expected_pkg_name
    assert result.path == tmp_path / expected_clean_name

    project_root = result.path
    module_dir = project_root / "my_cool_scanner"
    tests_dir = project_root / "tests"

    assert project_root.exists()
    assert module_dir.exists()
    assert tests_dir.exists()

    assert (project_root / "pyproject.toml").exists()
    assert (project_root / "README.md").exists()
    assert (project_root / ".gitignore").exists()
    assert (module_dir / "__init__.py").exists()
    assert (module_dir / "model.py").exists()
    assert (project_root / "model_config.toml").exists()
    assert (tests_dir / "test_model.py").exists()

    pyproject_content = (project_root / "pyproject.toml").read_text(encoding="utf-8")
    assert f'name = "{expected_pkg_name}"' in pyproject_content
    assert f'description = "Dorsal model: {project_name}"' in pyproject_content

    model_content = (module_dir / "model.py").read_text(encoding="utf-8")
    assert "class MyCoolScanner(AnnotationModel):" in model_content


def test_create_project_already_exists(tmp_path):
    """
    Test that we do not overwrite an existing directory.
    """

    project_name = "Existing Project"
    clean_name = "existing-project"
    (tmp_path / clean_name).mkdir()

    with pytest.raises(DorsalError) as exc:
        initialize.create_new_annotation_model_project(project_name, target_dir=tmp_path)

    assert f"Directory '{tmp_path / clean_name}' already exists" in str(exc.value)


def test_create_project_os_error(tmp_path):
    """
    Test that OS errors (like permission denied) are caught and re-raised as DorsalError.
    """

    with patch("pathlib.Path.mkdir", side_effect=OSError("Permission denied")):
        with pytest.raises(DorsalError) as exc:
            initialize.create_new_annotation_model_project("Fail Project", target_dir=tmp_path)

        assert "Failed to initialize project directory" in str(exc.value)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

PDFIUM_METADATA_FIELD_MAPPING = {
    "Creator": "creator",
    "Producer": "producer",
    "CreationDate": "creation_date",
    "Author": "author",
    "ModDate": "modified_date",
    "Title": "title",
    "Subject": "subject",
    "Keywords": "keywords",
    "version": "version",
    "page_count": "page_count",
}



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from pydantic import BaseModel, Field
from dorsal.common.validators import TString4096, TString256, TStringList256


class PDFValidationModel(BaseModel):
    """Validation model for core PDF metadata."""

    author: TString4096 | None = Field(
        default=None,
        description="The primary author(s) of the document, from the 'Author' metadata field.",
        json_schema_extra={"pii_risk": True},
    )
    title: TString4096 | None = Field(
        default=None, description="The title of the document, from the 'Title' metadata field."
    )
    creator: TString4096 | None = Field(
        default=None,
        description="The software tool used to create the original document (e.g., 'Word'), from the 'Creator' field.",
        json_schema_extra={"pii_risk": True},
    )
    producer: TString4096 | None = Field(
        default=None,
        description="The software tool used to convert or produce the final PDF (e.g., 'Acrobat PDF Library'), from the 'Producer' field.",
    )
    subject: TString4096 | None = Field(
        default=None, description="The subject or topic of the document, from the 'Subject' field."
    )
    keywords: TStringList256 = Field(
        default_factory=list, description="A list of keywords associated with the document."
    )
    version: TString256 | None = Field(
        default=None,
        description="The PDF version (e.g., '1.7', '2.0') as reported by the parser.",
    )
    page_count: int | None = Field(default=None, ge=0, description="The total number of pages in the document.")
    creation_date: datetime.datetime | None = Field(
        default=None, description="The date and time the document was created, from the 'CreationDate' field."
    )
    modified_date: datetime.datetime | None = Field(
        default=None, description="The date and time the document was last modified, from the 'ModDate' field."
    )



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import importlib.metadata
import importlib.resources
import tomllib
from typing import Any, Callable, cast

from packaging.utils import canonicalize_name

from dorsal.api.config import get_model_pipeline
from dorsal.common.exceptions import DorsalError, DorsalConfigError
from dorsal.common.validators import CallableImportPath
from dorsal.file.configs.model_runner import ModelRunnerPipelineStep, resolve_pipeline_step_models, RunModelResult
from dorsal.file.file_annotator import FILE_ANNOTATOR
from dorsal.file.model_runner import ModelRunner, run_model
from dorsal.file.validators.file_record import Annotation, AnnotationGroup
from dorsal.registry.installer import install_model_target
from dorsal.registry.resolution import resolve_target, is_package_installed

logger = logging.getLogger(__name__)

__all__ = ["run_or_install_model"]


def run_or_install_model(
    target: str,
    file_path: str,
    *,
    options: dict[str, Any] | None = None,
    ignore_linter_errors: bool = False,
    progress_callback: Callable[[float, float, str], None] | None = None,
) -> RunModelResult:
    """
    Resolve a model, installs if necessary, and execute it on a local file.

    Args:
        target: A Registry ID ('dorsalhub/whisper'), Package Name ('dorsal-whisper'),
                or Class Name ('FasterWhisperTranscriber').
        file_path: Path to the file to process.
        api_key: Optional API key.
        options: Runtime options to override the model's defaults.
        ignore_linter_errors: If True, bypasses data quality checks.

    Returns:
        Annotation | AnnotationGroup: The final, validated annotation object(s).
    """
    logger.info(f"Requesting run for model target '{target}' on file '{file_path}'")

    strategy, package_name = resolve_target(target)

    if not is_package_installed(package_name):
        if strategy == "registry_id":
            logger.info(f"Model package '{package_name}' is not installed. Installing from '{target}'...")
            try:
                install_model_target(target)
                logger.info(f"Successfully installed '{package_name}'.")
            except Exception as e:
                raise DorsalError(f"Failed to auto-install model '{target}': {e}") from e
        else:
            message = f"The model '{target}' is not installed locally.\n"
            if target == ".":
                message += "For local development, install the model before running it."
            raise DorsalError(message)

    pipeline_step = _get_execution_step(package_name)

    if options or ignore_linter_errors:
        pipeline_step = pipeline_step.model_copy(deep=True)
        if options:
            pipeline_step.options = {**(pipeline_step.options or {}), **options}
        if ignore_linter_errors:
            pipeline_step.ignore_linter_errors = True

    annotator_class, validator = resolve_pipeline_step_models(pipeline_step)

    if pipeline_step.schema_id and pipeline_step.schema_id.startswith("open/"):
        effective_validator = None
    else:
        effective_validator = validator

    # 2. Delegate directly to run_model
    run_result = run_model(
        annotation_model=annotator_class,
        file_path=file_path,
        schema_id=pipeline_step.schema_id,
        schema_version=pipeline_step.schema_version,
        validation_model=effective_validator,
        dependencies=pipeline_step.dependencies,
        options=pipeline_step.options,
        ignore_linter_errors=pipeline_step.ignore_linter_errors,
        progress_callback=progress_callback,
    )

    return run_result


def _get_execution_step(package_name: str) -> ModelRunnerPipelineStep:
    """Retrieves ModelRunnerPipelineStep for the given package."""

    pipeline = get_model_pipeline(scope="effective")
    safe_pkg_name = canonicalize_name(package_name)

    for step in pipeline:
        if step.package_name and canonicalize_name(step.package_name) == safe_pkg_name:
            logger.debug(f"Found existing pipeline configuration for '{package_name}'.")
            return step

    logger.debug(f"No pipeline config found for '{package_name}'. Constructing ephemeral step from package.")
    return _construct_step_from_package(package_name)


def _construct_step_from_package(package_name: str) -> ModelRunnerPipelineStep:
    """Orchestrate construction of a pipeline step from an installed package."""
    module_name = _resolve_module_from_package(package_name)
    config_data = _load_package_config(module_name, package_name)
    return _build_pipeline_step(config_data, module_name, package_name)


def _resolve_module_from_package(package_name: str) -> str:
    """Resolve module name for a package by inspecting 'dorsal.models' entry points."""
    safe_pkg_name = canonicalize_name(package_name)
    eps = importlib.metadata.entry_points(group="dorsal.models")

    for ep in eps:
        if ep.dist and canonicalize_name(ep.dist.name) == safe_pkg_name:
            return ep.module

    raise DorsalError(
        f"Package '{package_name}' is installed but does not declare a 'dorsal.models' entry point.\n"
        "This package may be broken or is not a compatible Dorsal model."
    )


def _load_package_config(module_name: str, package_name: str) -> dict[str, Any]:
    """Load and parse 'model_config.toml' from the package's module resources."""
    try:
        resource_path = importlib.resources.files(module_name) / "model_config.toml"
        if not resource_path.is_file():
            raise DorsalConfigError(
                f"Package '{package_name}' is missing 'model_config.toml' in module '{module_name}'"
            )

        config_text = resource_path.read_text(encoding="utf-8")
        return tomllib.loads(config_text)

    except (ImportError, FileNotFoundError) as err:
        raise DorsalConfigError(f"Could not load config for '{package_name}': {err}") from err
    except tomllib.TOMLDecodeError as err:
        raise DorsalConfigError(f"Syntax error in 'model_config.toml' for '{package_name}': {err}") from err


def _build_pipeline_step(config_data: dict[str, Any], module_name: str, package_name: str) -> ModelRunnerPipelineStep:
    """Validate configuration data and constructs the ModelRunnerPipelineStep object."""
    class_name = config_data.get("model_class")
    schema_id = config_data.get("schema_id")

    if not class_name:
        raise DorsalConfigError(f"Invalid config in '{package_name}': missing required field 'model_class'")

    if not schema_id:
        raise DorsalConfigError(f"Invalid config in '{package_name}': missing required field 'schema_id'")

    try:
        return ModelRunnerPipelineStep(
            annotation_model=CallableImportPath(module=module_name, name=class_name),
            schema_id=schema_id,
            schema_version=config_data.get("schema_version"),
            dependencies=config_data.get("dependencies"),
            validation_model=None,
            options=config_data.get("options"),
            package_name=package_name,
        )
    except Exception as e:
        raise DorsalError(f"Failed to construct pipeline step for '{package_name}': {e}") from e



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import requests

from unittest.mock import MagicMock
from dorsal.client import DorsalClient
from dorsal.common.exceptions import (
    DorsalClientError,
    BatchSizeError,
    ApiDataValidationError,
    APIError,
    NetworkError,
    APIError,
    ApiDataValidationError,
    ForbiddenError,
)
from dorsal.file.validators.file_record import (
    FileRecordStrict,
    FileRecordDateTime,
    FileSearchResponse,
    NewFileTag,
    ValidateTagsResult,
)
from dorsal.client.validators import FileIndexResponse, FileTagResponse, FileDeleteResponse, FileUrlResponse

_DUMMY_API_KEY = "abc123_test_key"
_DUMMY_BASE_URL = "http://dorsalhub.test"
_DUMMY_SHA256 = "a" * 64


@pytest.fixture
def client():
    return DorsalClient(api_key=_DUMMY_API_KEY, base_url=_DUMMY_BASE_URL)


@pytest.fixture
def mock_file_record_strict():
    return FileRecordStrict(
        hash=_DUMMY_SHA256,
        validation_hash="b" * 64,
        source="disk",
        annotations={
            "file/base": {
                "record": {
                    "hash": _DUMMY_SHA256,
                    "name": "test.txt",
                    "size": 100,
                    "media_type": "text/plain",
                    "all_hashes": [{"id": "SHA-256", "value": _DUMMY_SHA256}, {"id": "BLAKE3", "value": "b" * 64}],
                },
                "source": {"type": "Model", "id": "dorsal/file-core", "version": "1.0"},
            }
        },
    )


@pytest.fixture
def mock_file_response_json():
    """Standard JSON response for a downloaded file record."""
    return {
        "hash": _DUMMY_SHA256,
        "validation_hash": "b" * 64,
        "date_created": "2025-01-01T12:00:00Z",
        "date_modified": "2025-01-01T12:00:00Z",
        "annotations": {
            "file/base": {
                "record": {"hash": _DUMMY_SHA256, "name": "test.txt", "size": 100, "media_type": "text/plain"},
                "source": {"type": "Model", "id": "base", "version": "1.0"},
            }
        },
    }


def test_index_public_file_records_success(client, requests_mock, mock_file_record_strict):
    """Test successful indexing of public records."""
    mock_response = {"total": 1, "success": 1, "error": 0, "unauthorized": 0, "results": []}
    requests_mock.post(f"{_DUMMY_BASE_URL}/v1/files/public", json=mock_response, status_code=201)

    result = client.index_public_file_records([mock_file_record_strict])

    assert isinstance(result, FileIndexResponse)
    assert result.success == 1


def test_index_private_file_records_success(client, requests_mock, mock_file_record_strict):
    """Test successful indexing of private records."""
    mock_response = {"total": 1, "success": 1, "error": 0, "unauthorized": 0, "results": []}
    requests_mock.post(f"{_DUMMY_BASE_URL}/v1/files/private", json=mock_response, status_code=201)

    result = client.index_private_file_records([mock_file_record_strict])

    assert isinstance(result, FileIndexResponse)
    assert result.success == 1


def test_index_file_records_batch_size_error(client, mock_file_record_strict):
    """Test that exceeding batch size raises BatchSizeError."""
    client._file_records_batch_insert_size = 1
    records = [mock_file_record_strict, mock_file_record_strict]

    with pytest.raises(BatchSizeError):
        client.index_public_file_records(records)


def test_search_files_success(client, requests_mock):
    """Test successful file search."""
    mock_response = {
        "api_version": "1.0",
        "pagination": {
            "record_count": 1,
            "page_count": 1,
            "page": 1,
            "per_page": 25,
            "total_items": 1,
            "total_pages": 1,
            "current_page": 1,
            "has_next": False,
            "has_prev": False,
            "start_index": 0,
            "end_index": 0,
        },
        "results": [],
        "errors": [],
    }
    requests_mock.get(f"{_DUMMY_BASE_URL}/v1/files/search", json=mock_response, status_code=200)

    result = client.search_files(q="name:test", match_any=False)
    assert isinstance(result, FileSearchResponse)
    assert result.pagination.record_count == 1


def test_search_files_client_validation(client):
    """Test invalid search parameters."""
    with pytest.raises(DorsalClientError, match="'per_page' must be between"):
        client.search_files(q="test", per_page=1000, match_any=False)

    with pytest.raises(DorsalClientError, match="non-empty string"):
        client.search_files(q="", match_any=False)


def test_check_files_indexed(client, requests_mock):
    """Test bulk existence check."""
    requests_mock.post(
        f"{_DUMMY_BASE_URL}/v1/users/files-indexed", json={_DUMMY_SHA256: True, "other_hash": False}, status_code=200
    )

    valid_other_hash = "b" * 64
    requests_mock.post(
        f"{_DUMMY_BASE_URL}/v1/users/files-indexed",
        json={_DUMMY_SHA256: True, valid_other_hash: False},
        status_code=200,
    )

    result = client.check_files_indexed([_DUMMY_SHA256, valid_other_hash])
    assert result[_DUMMY_SHA256] is True
    assert result[valid_other_hash] is False


def test_validate_tag_success(client, requests_mock):
    """Test tag validation."""
    tag = NewFileTag(name="status", value="ok", private=False)
    requests_mock.post(
        f"{_DUMMY_BASE_URL}/v1/files/tags/validate", json={"valid": True, "message": "OK"}, status_code=200
    )

    result = client.validate_tag([tag])
    assert isinstance(result, ValidateTagsResult)
    assert result.valid is True


def test_add_tags_to_file_success(client, requests_mock):
    """Test adding tags."""
    tag = NewFileTag(name="status", value="ok", private=False)
    requests_mock.post(
        f"{_DUMMY_BASE_URL}/v1/files/{_DUMMY_SHA256}/tags",
        json={"success": True, "hash": _DUMMY_SHA256, "tags": []},
        status_code=200,
    )

    result = client.add_tags_to_file(_DUMMY_SHA256, [tag])
    assert isinstance(result, FileTagResponse)


def test_delete_tag_success(client, requests_mock):
    """Test deleting a tag."""
    tag_id = "tag_1"
    requests_mock.delete(f"{_DUMMY_BASE_URL}/v1/files/{_DUMMY_SHA256}/tags/{tag_id}", status_code=204)

    result = client.delete_tag(file_hash=_DUMMY_SHA256, tag_id=tag_id)
    assert result is None


def test_download_public_file_success(client, requests_mock, mock_file_response_json):
    """Test public download."""
    requests_mock.get(f"{_DUMMY_BASE_URL}/v1/files/public/{_DUMMY_SHA256}", json=mock_file_response_json)

    result = client.download_public_file_record(_DUMMY_SHA256)
    assert isinstance(result, FileRecordDateTime)
    assert result.hash == _DUMMY_SHA256


def test_download_private_file_success(client, requests_mock, mock_file_response_json):
    """Test private download."""
    requests_mock.get(f"{_DUMMY_BASE_URL}/v1/files/private/{_DUMMY_SHA256}", json=mock_file_response_json)

    result = client.download_private_file_record(_DUMMY_SHA256)
    assert isinstance(result, FileRecordDateTime)


def test_download_agnostic_fallback(client, requests_mock, mock_file_response_json):
    """Test that agnostic download tries private then public."""
    requests_mock.get(f"{_DUMMY_BASE_URL}/v1/files/private/{_DUMMY_SHA256}", status_code=404)
    requests_mock.get(f"{_DUMMY_BASE_URL}/v1/files/public/{_DUMMY_SHA256}", json=mock_file_response_json)

    result = client.download_file_record(_DUMMY_SHA256)
    assert result.hash == _DUMMY_SHA256
    assert requests_mock.call_count == 2


def test_download_invalid_response(client, requests_mock):
    """Test handling of malformed JSON from API."""
    requests_mock.get(
        f"{_DUMMY_BASE_URL}/v1/files/public/{_DUMMY_SHA256}", json={"incomplete": "data"}, status_code=200
    )

    with pytest.raises(ApiDataValidationError):
        client.download_public_file_record(_DUMMY_SHA256)


def test_delete_file_granular(client, requests_mock):
    """Test the granular delete endpoint."""
    requests_mock.delete(
        f"{_DUMMY_BASE_URL}/v1/files/{_DUMMY_SHA256}",
        json={"file_deleted": 1, "tags_deleted": 0, "annotations_deleted": 0},
        status_code=200,
    )

    result = client.delete_file(file_hash=_DUMMY_SHA256, record="private")
    assert isinstance(result, FileDeleteResponse)
    assert result.file_deleted == 1


def test_add_file_url_success(client, requests_mock):
    """Test successfully adding a URL to a file record using strict response validation."""
    target_url = "https://example.com/research-paper"

    mock_response = {"url_id": "url_abc123def456"}

    requests_mock.post(f"{_DUMMY_BASE_URL}/v1/files/{_DUMMY_SHA256}/urls", json=mock_response, status_code=201)

    result = client.add_file_url(file_hash=_DUMMY_SHA256, url=target_url, private=True, reference="test-ref")

    assert isinstance(result, FileUrlResponse)
    assert result.url_id == "url_abc123def456"

    last_request_json = requests_mock.last_request.json()
    assert "parent_url" not in last_request_json
    assert last_request_json["reference"] == "test-ref"
    assert last_request_json["url"] == target_url


def test_add_file_url_invalid_hash(client):
    """Test that adding a URL with an invalid hash raises a DorsalClientError."""
    with pytest.raises(DorsalClientError, match="Invalid SHA-256 hash provided"):
        client.add_file_url(file_hash="short-invalid-hash", url="https://example.com", private=False)


def test_delete_file_url_success(client, requests_mock):
    """Test successfully deleting a file URL."""
    url_id = "url_abc123def456"

    requests_mock.delete(f"{_DUMMY_BASE_URL}/v1/files/urls/{url_id}", status_code=204)

    result = client.delete_file_url(url_id=url_id)

    assert result is None
    assert requests_mock.called


def test_delete_file_url_invalid_id(client):
    """Test that deleting a URL with an empty string raises a DorsalClientError."""
    with pytest.raises(DorsalClientError, match="url_id must be a non-empty string"):
        client.delete_file_url(url_id="")


def test_add_file_url_network_error(client, requests_mock):
    """Test that a RequestException during add_file_url raises a NetworkError."""
    requests_mock.post(
        f"{_DUMMY_BASE_URL}/v1/files/{_DUMMY_SHA256}/urls",
        exc=requests.exceptions.RequestException("Simulated connection drop"),
    )

    with pytest.raises(NetworkError, match="An unexpected error occurred during the HTTP request."):
        client.add_file_url(file_hash=_DUMMY_SHA256, url="https://example.com", private=True)


def test_add_file_url_api_error(client, requests_mock):
    """Test that a non-201 response during add_file_url triggers _handle_api_error."""
    requests_mock.post(
        f"{_DUMMY_BASE_URL}/v1/files/{_DUMMY_SHA256}/urls", status_code=500, json={"detail": "Internal Server Error"}
    )

    with pytest.raises(APIError, match="Server error \\(500\\)"):
        client.add_file_url(file_hash=_DUMMY_SHA256, url="https://example.com", private=True)


def test_add_file_url_json_decode_error(client, requests_mock):
    """Test that malformed JSON response raises ApiDataValidationError."""
    requests_mock.post(
        f"{_DUMMY_BASE_URL}/v1/files/{_DUMMY_SHA256}/urls", status_code=201, text="<html>Not JSON</html>"
    )

    with pytest.raises(ApiDataValidationError, match="API response on add_file_url success failed data validation"):
        client.add_file_url(file_hash=_DUMMY_SHA256, url="https://example.com", private=True)


def test_add_file_url_pydantic_validation_error(client, requests_mock):
    """Test that a response missing required fields raises ApiDataValidationError."""
    requests_mock.post(
        f"{_DUMMY_BASE_URL}/v1/files/{_DUMMY_SHA256}/urls", status_code=201, json={"wrong_key": "missing_url_id"}
    )

    with pytest.raises(ApiDataValidationError, match="API response on add_file_url success failed data validation"):
        client.add_file_url(file_hash=_DUMMY_SHA256, url="https://example.com", private=True)


def test_delete_file_url_network_error(client, requests_mock):
    """Test that a RequestException during delete_file_url raises a NetworkError."""
    url_id = "url_abc123"
    requests_mock.delete(
        f"{_DUMMY_BASE_URL}/v1/files/urls/{url_id}",
        exc=requests.exceptions.RequestException("Simulated connection drop"),
    )

    with pytest.raises(NetworkError, match="An unexpected error occurred during the HTTP request."):
        client.delete_file_url(url_id=url_id)


def test_delete_file_url_api_error(client, requests_mock):
    """Test that a non-204 response during delete_file_url triggers _handle_api_error."""
    url_id = "url_abc123"
    requests_mock.delete(
        f"{_DUMMY_BASE_URL}/v1/files/urls/{url_id}",
        status_code=403,
        json={"detail": "You do not have permission to delete this URL."},
    )

    with pytest.raises(ForbiddenError, match="Forbidden: You do not have permission"):
        client.delete_file_url(url_id=url_id)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import csv
from collections import defaultdict
import datetime
import json
import logging
import operator
import os
import sqlite3
from typing import Any, Iterator, Sequence, TYPE_CHECKING, TypedDict, cast, Iterable

from dorsal.common.environment import is_jupyter_environment
from dorsal.file.dorsal_file import _DorsalFile
from dorsal.file.utils.size import human_filesize
from dorsal.version import __version__

if TYPE_CHECKING:
    from rich.console import Console

logger = logging.getLogger(__name__)


class _StatInfo(TypedDict):
    size: float
    path: str | None


class _DateStatInfo(TypedDict):
    date: datetime.datetime
    path: str | None


def _check_file_for_filter(file_obj: _DorsalFile, filter_criteria: dict) -> bool:
    """
    Checks if a single file object meets all specified filter criteria.

    Args:
        file_obj: The _DorsalFile object to check.
        filter_criteria: A dictionary of filter conditions to apply.

    Returns:
        True if the file meets all criteria, False otherwise.
    """
    op_map = {
        "gt": operator.gt,
        "lt": operator.lt,
        "gte": operator.ge,
        "lte": operator.le,
        "contains": lambda a, v: v in a if hasattr(a, "__contains__") else False,
        "in": lambda a, v: a in v if hasattr(v, "__contains__") else False,
    }

    for key, filter_value in filter_criteria.items():
        parts = key.split("__")
        attr_path_parts = parts[:-1] if len(parts) > 1 and parts[-1] in op_map else parts
        op_str = parts[-1] if len(parts) > 1 and parts[-1] in op_map else "exact"

        try:
            current_attr = file_obj
            for part in attr_path_parts:
                current_attr = getattr(current_attr, part)

            if op_str == "exact":
                if current_attr != filter_value:
                    return False
            else:
                op_func = op_map.get(op_str)
                if not op_func or not op_func(current_attr, filter_value):
                    return False
        except (AttributeError, TypeError):
            return False

    return True


class _BaseFileCollection:
    """Base class providing shared functionality for file collections."""

    def __init__(self, files: Sequence[_DorsalFile], source_info: dict | None = None):
        """
        Args:
            files: A list of objects that inherit from _DorsalFile.
            source_info: Metadata about how the collection was created.
        """
        self.files: Sequence[_DorsalFile] = files
        self.source_info: dict = source_info or {}
        self._is_populated: bool = False

    def __len__(self) -> int:
        """Returns the number of files in the collection."""
        return len(self.files)

    def __iter__(self) -> Iterator[_DorsalFile]:
        """Returns an iterator over the files in the collection."""
        return iter(self.files)

    def __getitem__(self, index: int) -> _DorsalFile | list[_DorsalFile]:
        """Allows accessing files by index."""
        return self.files[index]

    def __repr__(self) -> str:
        """Provides a developer-friendly string representation of the collection."""
        source_type = self.source_info.get("type", "generic")

        if source_type == "local":
            path = self.source_info.get("path", "from list")
            display_path = f"...{path[-22:]}" if len(path) > 25 else path
        elif source_type == "remote":
            collection_id = self.source_info.get("collection_id")
            display_path = f"remote id: {collection_id}"
        else:
            display_path = "from list"

        return f"<{self.__class__.__name__} [{display_path}] ({len(self)})>"

    def info(self) -> dict:
        """
        Provides a summary of the file collection.
        """
        if not self._is_populated:
            raise TypeError(
                "The 'stats' method can only be called on a fully populated collection. "
                "Please call the .populate() method first."
            )

        if not self.files:
            return {
                "overall": {
                    "total_files": 0,
                    "total_size": 0,
                    "avg_size": 0,
                    "largest_file": None,
                    "smallest_file": None,
                    "newest_file": None,
                    "oldest_file": None,
                },
                "by_type": [],
            }

        total_size = 0
        smallest: _StatInfo = {"size": float("inf"), "path": ""}
        largest: _StatInfo = {"size": float("-inf"), "path": ""}
        oldest: _DateStatInfo = {
            "date": datetime.datetime.now().astimezone(),
            "path": "",
        }
        newest: _DateStatInfo = {
            "date": datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc),
            "path": "",
        }
        type_stats: defaultdict[str, dict[str, int]] = defaultdict(lambda: {"count": 0, "total_size": 0})
        source_stats: defaultdict[str, dict[str, int]] = defaultdict(lambda: {"count": 0, "total_size": 0})

        for f in self.files:
            mod_time = f.date_modified
            total_size += f.size

            if f.size < smallest["size"]:
                smallest["size"] = f.size
                smallest["path"] = f.name
            if f.size > largest["size"]:
                largest["size"] = f.size
                largest["path"] = f.name
            if mod_time < oldest["date"]:
                oldest["date"] = mod_time
                oldest["path"] = f.name
            if mod_time > newest["date"]:
                newest["date"] = mod_time
                newest["path"] = f.name

            media_type = f.media_type or "unknown"
            type_stats[media_type]["count"] += 1
            type_stats[media_type]["total_size"] += f.size

            source = f._source
            source_stats[source]["count"] += 1
            source_stats[source]["total_size"] += f.size

        type_breakdown = [
            {
                "media_type": mt,
                "count": data["count"],
                "total_size": data["total_size"],
                "percentage_of_total": ((data["total_size"] / total_size) * 100 if total_size > 0 else 0),
            }
            for mt, data in type_stats.items()
        ]
        type_breakdown.sort(key=lambda x: cast(int, x["total_size"]), reverse=True)

        source_breakdown = [
            {
                "source": source,
                "count": data["count"],
                "total_size": data["total_size"],
                "percentage_of_total": ((data["total_size"] / total_size) * 100 if total_size > 0 else 0),
            }
            for source, data in source_stats.items()
        ]
        source_breakdown.sort(key=lambda x: cast(int, x["total_size"]), reverse=True)

        return {
            "overall": {
                "total_files": len(self.files),
                "total_size": total_size,
                "avg_size": total_size / len(self.files) if self.files else 0,
                "largest_file": largest,
                "smallest_file": smallest,
                "newest_file": newest,
                "oldest_file": oldest,
            },
            "by_type": type_breakdown,
            "by_source": source_breakdown,
        }

    def find_duplicates(
        self,
        min_size_bytes: int = 0,
        max_size_bytes: int | None = None,
        console: "Console | None" = None,
        palette: dict | None = None,
    ) -> dict:
        """
        Finds sets of files with identical content hashes in the collection.

        Args:
            min_size_bytes (int): Minimum file size to include in the search.
            max_size_bytes (int | None): Maximum file size to include.
            console (Console | None): Rich Console object for progress display.
            palette (dict | None): Color palette for styling the progress bar.
        """
        if not self._is_populated:
            raise TypeError(
                "The 'find_duplicates' method can only be called on a fully populated collection. "
                "Please call the .populate() method first."
            )
        if not self.files:
            return {}

        rich_progress = None
        files_to_check = [
            f for f in self.files if f.size > min_size_bytes and (max_size_bytes is None or f.size <= max_size_bytes)
        ]

        iterator: Iterable[_DorsalFile]
        if is_jupyter_environment():
            from tqdm import tqdm

            iterator = tqdm(files_to_check, desc="Hashing files")
        elif console:
            from rich.progress import (
                Progress,
                BarColumn,
                TaskProgressColumn,
                MofNCompleteColumn,
                TextColumn,
                TimeElapsedColumn,
                TimeRemainingColumn,
            )
            from dorsal.cli.themes.palettes import DEFAULT_PALETTE

            active_palette = palette if palette is not None else DEFAULT_PALETTE
            progress_columns = (
                TextColumn(
                    "[progress.description]{task.description}",
                    style=active_palette.get("progress_description", "default"),
                ),
                BarColumn(bar_width=None, style=active_palette.get("progress_bar", "default")),
                TaskProgressColumn(style=active_palette.get("progress_percentage", "default")),
                MofNCompleteColumn(),
                TextColumn("•", style="dim"),
                TimeElapsedColumn(),
                TextColumn("•", style="dim"),
                TimeRemainingColumn(),
            )
            rich_progress = Progress(
                *progress_columns,
                console=console,
                redirect_stdout=True,
                redirect_stderr=True,
            )
            task_id = rich_progress.add_task("Finding duplicates...", total=len(files_to_check))
            iterator = files_to_check
        else:
            iterator = files_to_check

        hash_map = defaultdict(list)
        with rich_progress if rich_progress else open(os.devnull, "w"):
            for file in iterator:
                hash_map[file.hash].append(file)
                if rich_progress:
                    rich_progress.update(task_id, advance=1)

        duplicate_sets_raw: list[list[_DorsalFile]] = [files for files in hash_map.values() if len(files) > 1]
        if not duplicate_sets_raw:
            return {}

        total_wasted_space = 0
        duplicate_sets_formatted = []
        for files in duplicate_sets_raw:
            first_file = files[0]
            count = len(files)
            size_each = first_file.size
            wasted_for_set = size_each * (count - 1)
            total_wasted_space += wasted_for_set

            paths = [getattr(f, "_file_path", f.name) for f in files]

            duplicate_sets_formatted.append(
                {
                    "hash": first_file.hash,
                    "count": count,
                    "file_size": human_filesize(size_each),
                    "file_size_bytes": size_each,
                    "wasted_space": human_filesize(wasted_for_set),
                    "wasted_space_bytes": wasted_for_set,
                    "paths": paths,
                }
            )

        duplicate_sets_formatted.sort(key=lambda x: cast(int, x["wasted_space_bytes"]), reverse=True)

        return {
            "total_sets": len(duplicate_sets_formatted),
            "total_wasted_space": human_filesize(total_wasted_space),
            "total_wasted_space_bytes": total_wasted_space,
            "duplicate_sets": duplicate_sets_formatted,
        }

    def filter(self, **kwargs: Any):
        """
        Filters the collection based on file attributes using flexible criteria.
        ...
        """
        if not kwargs:
            return self.__class__(files=list(self.files), source_info=self.source_info)

        filtered_files = [f for f in self.files if _check_file_for_filter(f, kwargs)]

        return self.__class__(files=filtered_files, source_info=self.source_info)

    def _get_flattened_data(self) -> tuple[list[str], list[dict]]:
        """
        Flattens the collection's data for tabular export formats.
        ...
        """
        header_set = set()
        rows = []
        for file in self.files:
            row = {
                "source_path": self.source_info.get("path"),
                "hash": file.hash,
                "file_path": getattr(file, "_file_path", None),
            }
            dumped_model = file.to_dict()
            annotations: dict[str, Any] = dumped_model.get("annotations", {})

            if not annotations:
                rows.append(row)
                continue

            for key, annotation_data in annotations.items():
                if annotation_data and isinstance(annotation_data, dict) and annotation_data.get("record"):
                    record: dict[str, Any] = annotation_data["record"]
                    prefix = key.replace("/", "_")
                    for field, value in record.items():
                        if isinstance(value, (str, int, float, bool)) or value is None:
                            col_name = f"{prefix}__{field}"
                            header_set.add(col_name)
                            row[col_name] = value
            rows.append(row)
        header = ["hash", "file_path", "source_path"] + sorted(list(header_set))
        return header, rows

    def to_csv(self, file_path: str) -> None:
        """Exports the collection's metadata to a CSV file."""
        if not self.files:
            return
        headers, rows = self._get_flattened_data()
        if not rows:
            return
        with open(file=file_path, mode="w", newline="", encoding="utf-8") as f:
            writer = csv.DictWriter(f, fieldnames=headers)
            writer.writeheader()
            writer.writerows(rows)
        logger.info(f"Successfully exported {len(rows)} records to {file_path}")

    def to_dataframe(self):  # pragma: no cover
        """Exports the collection's metadata to a pandas DataFrame."""
        try:
            import pandas as pd  # type: ignore
        except ImportError as err:
            raise ImportError("To use the `to_dataframe` method you must install pandas: `pip install pandas`") from err

        headers, rows = self._get_flattened_data()
        if not rows:
            return pd.DataFrame()
        return pd.DataFrame(data=rows, columns=headers)

    def to_dict(
        self,
        by_alias: bool = True,
        exclude_none: bool = True,
        exclude: dict | set | None = None,
    ) -> dict:
        """
        Serializes the collection to a dictionary.
        ...
        """
        collection_info = self.info()

        scan_metadata = {
            **self.source_info,
            "total_files_in_collection": len(self),
            "dorsal_version": __version__,
            "overall": collection_info.get("overall", {}),
            "by_type": collection_info.get("by_type", []),
            "by_source": collection_info.get("by_source", []),
        }

        return {
            "scan_metadata": scan_metadata,
            "results": [
                f.to_dict(
                    mode="json",
                    by_alias=by_alias,
                    exclude_none=exclude_none,
                    exclude=exclude,
                )
                for f in self.files
            ],
        }

    def to_json(
        self,
        filepath: str | None = None,
        indent: int | None = 2,
        by_alias: bool = True,
        exclude_none: bool = True,
        exclude: dict | set | None = None,
    ) -> str | None:
        """Saves the collection data to a structured JSON file or returns it as a string."""
        output_data = self.to_dict(by_alias=by_alias, exclude_none=exclude_none, exclude=exclude)

        def custom_encoder(obj):
            if isinstance(obj, datetime.datetime):
                return obj.isoformat()
            return str(obj)

        if filepath:
            with open(filepath, "w", encoding="utf-8") as f:
                json.dump(output_data, f, indent=indent, default=str)
            logger.info(f"Successfully exported {len(self)} records to {filepath}")
            return None
        else:
            return json.dumps(output_data, indent=indent, default=custom_encoder)

    def to_sqlite(self, db_path: str, table_name: str = "files") -> None:
        """Exports the collection's data to a table in an SQLite database."""
        if not self.files:
            logger.debug("FileCollection is empty, skipping SQLite export.")
            return

        headers, rows = self._get_flattened_data()
        if not rows:
            logger.debug("No data to export to SQLite.")
            return

        sanitized_headers = [h.replace("-", "_") for h in headers]

        conn = None
        try:
            conn = sqlite3.connect(db_path)
            cursor = conn.cursor()

            cols_with_types = ", ".join(f'"{col}" TEXT' for col in sanitized_headers)
            create_table_sql = f'CREATE TABLE IF NOT EXISTS "{table_name}" ({cols_with_types})'
            cursor.execute(create_table_sql)

            placeholders = ", ".join(["?"] * len(sanitized_headers))
            insert_sql = (
                f'INSERT INTO "{table_name}" ({", ".join(f"{h}" for h in sanitized_headers)}) VALUES ({placeholders})'
            )

            rows_to_insert = [tuple(row.get(h) for h in headers) for row in rows]

            cursor.executemany(insert_sql, rows_to_insert)
            conn.commit()
            logger.info(f"Successfully exported {len(rows)} records to table '{table_name}' in {db_path}")

        except sqlite3.Error as err:
            logger.exception(f"An error occurred during SQLite export to {db_path}")
            raise err
        finally:
            if conn:
                conn.close()



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dorsal.api.file import (
    add_tag_to_file,
    delete_private_dorsal_file_record,
    delete_public_dorsal_file_record,
    find_duplicates,
    generate_html_directory_report,
    generate_html_file_report,
    get_directory_info,
    get_dorsal_file_record,
    identify_file,
    index_directory,
    index_file,
    scan_directory,
    scan_file,
    remove_tag_from_file,
    search_global_files,
    search_user_files,
)

from dorsal.api.dataset import (
    get_dataset,
    get_dataset_schema,
    make_schema_validator,
    validate_dataset_records,
)

from dorsal.api.collection import (
    add_files_to_collection,
    delete_collection,
    export_collection,
    get_collection,
    list_collections,
    make_collection_private,
    make_collection_public,
    remove_files_from_collection,
    update_collection,
)

from dorsal.api.config import (
    register_model,
    show_model_pipeline,
    remove_model_by_name,
    activate_model_by_name,
    deactivate_model_by_name,
)


__all__ = [
    # file
    "add_tag_to_file",
    "delete_private_dorsal_file_record",
    "delete_public_dorsal_file_record",
    "find_duplicates",
    "generate_html_directory_report",
    "generate_html_file_report",
    "get_directory_info",
    "get_dorsal_file_record",
    "identify_file",
    "index_directory",
    "index_file",
    "scan_directory",
    "scan_file",
    "remove_tag_from_file",
    "search_global_files",
    "search_user_files",
    # dataset
    "get_dataset",
    "get_dataset_schema",
    "make_schema_validator",
    "validate_dataset_records",
    # collection
    "add_files_to_collection",
    "delete_collection",
    "export_collection",
    "get_collection",
    "list_collections",
    "make_collection_private",
    "make_collection_public",
    "remove_files_from_collection",
    "update_collection",
    # config
    "register_model",
    "show_model_pipeline",
    "remove_model_by_name",
    "activate_model_by_name",
    "deactivate_model_by_name",
]



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import typer
import json
from typing import Annotated
from rich.panel import Panel
from rich.text import Text
import logging
import enum

logger = logging.getLogger(__name__)


class DeletionScope(str, enum.Enum):
    all = "all"
    public = "public"
    private = "private"
    none = "none"


def delete_file_record(
    ctx: typer.Context,
    hash_string: Annotated[str, typer.Argument(help="The hash of the file record to delete.")],
    record: Annotated[
        DeletionScope,
        typer.Option(
            case_sensitive=False,
            help="Specify which core file record(s) to delete.",
        ),
    ] = DeletionScope.all,
    tags: Annotated[
        DeletionScope,
        typer.Option(
            case_sensitive=False,
            help="Specify which of the user's tags to delete for this hash.",
        ),
    ] = DeletionScope.all,
    annotations: Annotated[
        DeletionScope,
        typer.Option(
            case_sensitive=False,
            help="Specify which of the user's annotations to delete for this hash.",
        ),
    ] = DeletionScope.all,
    yes: Annotated[
        bool,
        typer.Option(
            "--yes",
            "-y",
            help="Bypass the interactive confirmation prompt. Use with caution.",
        ),
    ] = False,
    json_output: Annotated[bool, typer.Option("--json", help="Output result as a raw JSON object.")] = False,
):
    """
    Deletes a file record and/or its associated data from DorsalHub with granular control.

    The default behavior is a "full clean", targeting all records, tags, and
    annotations associated with the user for the given hash.
    """
    from dorsal.api.file import get_dorsal_file_record, _delete_dorsal_file_record
    from dorsal.cli.views.file import create_file_info_panel
    from dorsal.common.cli import get_rich_console, exit_cli, handle_error
    from dorsal.common.exceptions import AuthError, DorsalClientError, DorsalOfflineError, NotFoundError

    console = get_rich_console()
    palette: dict[str, str] = ctx.obj["palette"]
    hash_color = palette.get("hash_value", "default")

    fetch_scope: bool | None
    if record == DeletionScope.private:
        fetch_scope = False
    elif record == DeletionScope.public:
        fetch_scope = True
    else:
        fetch_scope = None

    try:
        if not yes and not json_output:
            console.print("🔎 Finding record to delete...")
            file_to_delete = get_dorsal_file_record(hash_string=hash_string, mode="pydantic", public=fetch_scope)
            record_dict = file_to_delete.model_dump(by_alias=True, exclude_none=True, mode="json")

            panel = create_file_info_panel(
                record_dict=record_dict,
                title="File record to be delete",
                private=fetch_scope,
                palette=palette,
                override_title_style=palette.get("error", "default"),
                override_border_style=palette.get("panel_border_error", "default"),
            )
            console.print(panel)

            bullet_points = []

            if record == DeletionScope.all:
                bullet_points.append("- Delete both [bold]public and private[/] records")
            elif record == DeletionScope.public:
                bullet_points.append("- Delete the [bold]public[/] record")
            elif record == DeletionScope.private:
                bullet_points.append("- Delete the [bold]private[/] record")

            if tags == DeletionScope.all:
                bullet_points.append("- Delete your [bold]public and private[/] tags")
            elif tags == DeletionScope.public:
                bullet_points.append("- Delete your [bold]public[/] tags")
            elif tags == DeletionScope.private:
                bullet_points.append("- Delete your [bold]private[/] tags")

            if annotations == DeletionScope.all:
                bullet_points.append("- Delete your [bold]public and private[/] annotations")
            elif annotations == DeletionScope.public:
                bullet_points.append("- Delete your [bold]public[/] annotations")
            elif annotations == DeletionScope.private:
                bullet_points.append("- Delete your [bold]private[/] annotations")

            if not bullet_points:
                summary_text = "This operation will perform no deletions."
            else:
                header = f"For the hash [{hash_color}]{hash_string}...[/{hash_color}] this operation will:"
                body = "\n".join(bullet_points)
                summary_text = f"{header}\n{body}"

            console.print(
                Panel(
                    Text.from_markup(summary_text, justify="left"),
                    title="Action Summary",
                    border_style=palette.get("panel_border_warning", "yellow"),
                    padding=(1, 2),
                    expand=False,
                )
            )

            console.print(
                f"[{palette.get('warning', 'default')}]Are you sure you want to perform this deletion operation?[/] ",
                end="",
            )
            typer.confirm("", abort=True, show_default=True)

        if not json_output:
            console.print(f"🗑️ Deleting record [{palette.get('hash_value', 'magenta')}]{hash_string}[/]")

        response = _delete_dorsal_file_record(
            file_hash=hash_string, record=record.value, tags=tags.value, annotations=annotations.value
        )

        if json_output:
            console.print(response.model_dump_json(indent=2))
        else:
            if response.file_deleted == 0 and response.file_modified == 0:
                message = "Operation completed. No records were deleted or modified."
            elif response.file_deleted > 0 and response.file_modified > 0:
                message = f"{response.file_deleted} record(s) deleted and {response.file_modified} modified."
            elif response.file_deleted > 0:
                message = f"{response.file_deleted} record(s) permanently deleted."
            else:
                message = "Public file record ownership rolled back."

            details = f"Removed {response.tags_deleted} tag(s) and {response.annotations_deleted} annotation(s)."

            success_text = Text.assemble(
                (f"{message}\n\n", palette.get("success", "bold green")),
                (details, palette.get("key", "dim")),
            )
            console.print(
                Panel(
                    success_text,
                    expand=False,
                    title="Deletion Complete",
                    border_style=palette.get("panel_border_success", "green"),
                )
            )

    except NotFoundError:
        scope_str = "any of your accessible records"
        if fetch_scope is True:
            scope_str = "your private records"
        elif fetch_scope is False:
            scope_str = "public records"

        message = f"File record with hash '{hash_string}' not found in {scope_str}."
        handle_error(palette, f"Cannot delete: {message}", json_output)
    except typer.Abort:
        if not json_output:
            console.print(f"[{palette.get('warning', 'default')}]Deletion aborted by user.[/]")
        exit_cli()
    except DorsalOfflineError:
        raise
    except AuthError:
        raise
    except DorsalClientError as e:
        handle_error(palette, f"API Error: {e.message}", json_output)
    except Exception as e:
        logger.exception("An unexpected error occurred during 'file delete'")
        handle_error(palette, f"An unexpected error occurred: {e}", json_output)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import time
import pathlib
import datetime
import re
from typing import Annotated, Literal, Optional

from rich.panel import Panel
from rich.table import Table
from rich.markup import escape
import typer

logger = logging.getLogger(__name__)


DuplicateMode = Literal["hybrid", "quick", "sha256"]


def duplicates_dir(
    ctx: typer.Context,
    path: Annotated[
        pathlib.Path,
        typer.Argument(
            exists=True,
            file_okay=False,
            dir_okay=True,
            readable=True,
            help="The directory path to scan for duplicates.",
        ),
    ],
    recursive: Annotated[
        bool,
        typer.Option(
            "--recursive/--no-recursive",
            "-r/-R",
            help="Scan subdirectories recursively.",
            rich_help_panel="Scan Options",
        ),
    ] = False,
    limit: Annotated[
        int,
        typer.Option(
            "--limit",
            "-l",
            help="The number of duplicate sets to display.",
            rich_help_panel="Scan Options",
        ),
    ] = 5,
    min_size: Annotated[
        str,
        typer.Option(
            "--min-size",
            help="Only find duplicates larger than this size (e.g., '1MB', '500KB').",
            rich_help_panel="Scan Options",
        ),
    ] = "0",
    max_size: Annotated[
        str | None,
        typer.Option(
            "--max-size",
            help="Only find duplicates smaller than this size (e.g., '1GB').",
            rich_help_panel="Scan Options",
        ),
    ] = None,
    output_path: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            "--output",
            "-o",
            help="Path to save the full JSON report (e.g., 'duplicates.json').",
            dir_okay=True,
            file_okay=True,
            writable=True,
            resolve_path=True,
            rich_help_panel="Output Options",
        ),
    ] = None,
    save: Annotated[
        bool,
        typer.Option(
            "-s",
            "--save",
            help="Save the full JSON report to the default directory or --output path.",
            rich_help_panel="Output Options",
        ),
    ] = False,
    json_output: Annotated[
        bool,
        typer.Option(
            "--json",
            help="Output the full duplicate report as a raw JSON object to stdout for scripting.",
            rich_help_panel="Output Options",
        ),
    ] = False,
    hybrid: Annotated[
        bool,
        typer.Option(
            "--hybrid",
            help="Use hybrid mode (default): fast initial scan with definitive SHA-256 verification.",
            rich_help_panel="Hashing Options",
        ),
    ] = False,
    quick: Annotated[
        bool,
        typer.Option(
            "--quick",
            "-q",
            help="Use QUICK hash only. Fastest, but may have rare false positives.",
            rich_help_panel="Hashing Options",
        ),
    ] = False,
    sha256: Annotated[
        bool,
        typer.Option(
            "--sha256",
            help="Use SHA-256 hash only. Slower, but provides definitive results.",
            rich_help_panel="Hashing Options",
        ),
    ] = False,
    use_cache: Annotated[
        bool,
        typer.Option(
            "--use-cache",
            help="Force the use of the cache, overriding any global setting.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    skip_cache: Annotated[
        bool,
        typer.Option(
            "--skip-cache",
            help="Bypass the local cache and re-calculate all hashes, overriding any global setting to enable it.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
):
    """
    Finds and reports files with identical content hashes.
    """
    from dorsal.api.file import find_duplicates
    from dorsal.common.cli import (
        EXIT_CODE_ERROR,
        get_rich_console,
        exit_cli,
        determine_use_cache_value,
    )

    console = get_rich_console()
    palette: dict[str, str] = ctx.obj["palette"]
    progress_console = None if json_output else console

    start_time = time.perf_counter()

    if use_cache and skip_cache:
        exit_cli(
            code=EXIT_CODE_ERROR,
            message="Error: --use-cache and --skip-cache flags cannot be used together.",
        )

    if output_path and not save:
        if str(output_path).lower().endswith(".json"):
            save = True
        else:
            if not output_path.is_dir():
                console.print(
                    f"⚠️ [yellow]Warning:[/] --output path '{output_path}' was specified with an unknown extension."
                    f" Please use -s (for .json) or specify a .json file.",
                    style=palette.get("warning", "yellow"),
                )

    use_cache_value = determine_use_cache_value(use_cache=use_cache, skip_cache=skip_cache)

    mode_flags = [hybrid, quick, sha256]
    if sum(mode_flags) > 1:
        exit_cli(
            code=EXIT_CODE_ERROR,
            message="Please specify only one hashing mode: --hybrid, --quick, or --sha256.",
        )

    mode: DuplicateMode = "hybrid"
    if quick:
        mode = "quick"
    elif sha256:
        mode = "sha256"

    if not json_output:
        console.print(f"Searching for duplicate files in [{palette['primary_value']}]{escape(str(path))}[/]")
        if mode == "quick":
            console.print(
                f"[{palette['info']}]Mode: Quick Hash. Results are based on file sampling and may contain rare false positives.[/]"
            )
        elif mode == "sha256":
            console.print(
                f"[{palette['info']}]Mode: Secure Hash. All files will be fully read for maximum accuracy.[/]"
            )
        else:
            console.print(f"[{palette['info']}]Mode: Hybrid. Using fast scan with secure verification.[/]")

    try:
        results = find_duplicates(
            path=path,
            recursive=recursive,
            min_size=min_size,
            max_size=max_size,
            mode=mode,
            progress_console=progress_console,
            palette=palette,
            use_cache=use_cache_value,
        )

        duration = time.perf_counter() - start_time

        results["duration_seconds"] = duration
        if "total_sets" not in results:
            results["total_sets"] = 0
        if "duplicate_sets" not in results:
            results["duplicate_sets"] = []
        if "path" not in results:
            results["path"] = str(path)

        if json_output:
            console.print(json.dumps(results, indent=2, default=str, ensure_ascii=False))
            exit_cli()

        if not results or not results.get("duplicate_sets"):
            console.print(f"\n[{palette['success']}]✅ No duplicate files found in {duration:.3f} seconds.[/]")
            exit_cli()

        total_sets = results["total_sets"]
        hashes_from_cache = results.get("hashes_from_cache", 0)
        cache_info_str = (
            f" ([{palette.get('success', 'green')}]{hashes_from_cache} hashes from cache[/])"
            if hashes_from_cache > 0
            else ""
        )
        console.print(
            f"\n⚠️ Found [{palette['warning']}]{total_sets}[/] set(s) of duplicate files{cache_info_str} in {duration:.3f} seconds."
        )

        if total_sets > 0 and limit > 0:
            console.print(f"\nDisplaying the {min(limit, total_sets)} largest duplicate sets.")

        sets_to_display = results["duplicate_sets"][:limit]

        for i, dupe_set in enumerate(sets_to_display):
            size_str = dupe_set["file_size"]
            hash_str = dupe_set.get("hash", "N/A")
            hash_type = dupe_set.get("hash_type", "N/A").upper()

            title_base = f"Duplicate Set {i + 1} of {total_sets}"
            size_info = f"Size: [{palette['success']}]{size_str}[/] (each)"
            title = f"{title_base} | {size_info}"

            hash_info_body = f"[{palette['info']}]{hash_type}: {hash_str}[/]"

            paths_table = Table.grid(padding=(0, 1, 0, 3))
            paths_table.add_column(style=palette["primary_value"])

            for p in dupe_set["paths"]:
                path_obj = pathlib.Path(p)
                path_display = f"📄 {escape(p)}"

                if path_obj.is_symlink():
                    try:
                        target = path_obj.readlink()
                        path_display += f" [dim italic]→ {escape(str(target))}[/]"
                    except OSError:
                        path_display += " [dim italic](symlink)[/]"

                paths_table.add_row(path_display)

            content_grid = Table.grid(padding=(0, 0, 1, 0))
            content_grid.add_row(hash_info_body)
            content_grid.add_row(paths_table)

            console.print(
                Panel(
                    content_grid,
                    title=title,
                    expand=False,
                    border_style=palette["panel_border"],
                )
            )

        if total_sets > limit and limit > 0:
            console.print(f"\n... and {total_sets - limit} more sets.")

        if save:
            _save_duplicates_report(
                results=results,
                output_path=output_path,
                palette=palette,
                json_output=json_output,
                original_path=path,
            )

    except typer.Exit:
        raise
    except Exception as err:
        logger.exception("CLI 'duplicates' command failed.")
        console_in_handler = get_rich_console()
        logger.debug(f"DEBUG: Console ID in error handler: {id(console_in_handler)}")
        exit_cli(code=EXIT_CODE_ERROR, message=str(err))


def _sanitize_path_for_filename(path: str) -> str:
    """Sanitizes a path to be used in a filename."""
    sanitized = re.sub(r'[<>:"/\\|?*]', "_", path)
    sanitized = sanitized.strip("_ ")

    return sanitized[:50]


def _get_final_path(source_path: pathlib.Path, output_path: Optional[pathlib.Path], suffix: str) -> pathlib.Path:
    """Helper to determine the final save path for a report."""
    from dorsal.common import constants

    if output_path:
        if output_path.is_dir():
            dir_name = _sanitize_path_for_filename(str(source_path.name))
            return output_path / f"{dir_name}{suffix}"
        else:
            return output_path

    constants.CLI_DUPLICATES_REPORTS_DIR.mkdir(parents=True, exist_ok=True)
    timestamp = datetime.datetime.now().strftime("%Y%m%d-%HM%S")
    dir_name = _sanitize_path_for_filename(str(source_path.name))
    return constants.CLI_DUPLICATES_REPORTS_DIR / f"{dir_name}_{timestamp}{suffix}"


def _save_duplicates_report(
    results: dict,
    output_path: Optional[pathlib.Path],
    palette: dict,
    json_output: bool,
    original_path: pathlib.Path,
):
    from dorsal.common.cli import get_rich_console

    console = get_rich_console()

    save_path = _get_final_path(original_path, output_path, ".json")

    try:
        save_path.parent.mkdir(parents=True, exist_ok=True)
        with open(save_path, "w", encoding="utf-8") as f:
            json.dump(results, f, indent=2, ensure_ascii=False)

        if not json_output:
            console.print(
                f"[{palette['success']}]✅ Full JSON report saved to:[/] [{palette['primary_value']}]{save_path}[/]"
            )
    except IOError as e:
        logger.error(f"Failed to save duplicates report: {e}")
        if not json_output:
            console.print(f"\n[{palette['error']}]Warning:[/] Could not save report to {save_path}. Error: {e}")



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
import zipfile
import xml.etree.ElementTree as ET
from typing import Any

from .utils import (
    RELS_NS,
    CORE_PROPS_TYPE,
    APP_PROPS_TYPE,
    DEFAULT_CORE_PROPS_PATH,
    DEFAULT_APP_PROPS_PATH,
    BASE_NS_MAP,
    _find_relationship_target,
    _read_xml_from_zip,
    _parse_core_properties,
    _parse_app_properties,
    _parse_content_types_xml,
    _parse_custom_properties_xml,
    _find_xml_element,
    _findall_xml_elements,
)


logger = logging.getLogger(__name__)

R_NS = "{http://schemas.openxmlformats.org/officeDocument/2006/relationships}"

XLSX_MAIN_NS = "{http://schemas.openxmlformats.org/spreadsheetml/2006/main}"

NS_MAP = BASE_NS_MAP.copy()
NS_MAP.update(
    {
        "m": "http://schemas.openxmlformats.org/spreadsheetml/2006/main",
    }
)

WORKSHEET_TYPE = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/worksheet"
SHARED_STRINGS_TYPE = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/sharedStrings"

CELL_REF_RX = re.compile(r"([A-Z]+)(\d+)")


def _col_letter_to_int(col_str: str) -> int:
    """Converts a column letter (A, B, ... Z, AA, AB) to its 1-based index."""
    num = 0
    for c in col_str:
        num = num * 26 + (ord(c.upper()) - ord("A")) + 1
    return num


def _parse_dimension_ref(ref_str: str | None) -> tuple[int | None, int | None]:
    """Parses a dimension string (e.g., 'A1:F10' or 'C5') into (row_count, col_count)."""
    if not ref_str:
        return None, None

    if ":" in ref_str:
        _, end_ref = ref_str.split(":")
    else:
        end_ref = ref_str

    match = CELL_REF_RX.match(end_ref)
    if not match:
        return None, None

    try:
        col_str = match.group(1)
        row_str = match.group(2)
        return int(row_str), _col_letter_to_int(col_str)
    except Exception as e:
        logger.warning(f"Failed to parse dimension reference: '{end_ref}': {e}")
        return None, None


def _get_cell_value(cell_el: ET.Element | None, shared_strings: list[str]) -> Any | None:
    """Gets the value from a <c> cell element, looking up shared strings if needed."""
    if cell_el is None:
        return None

    val_el = _find_xml_element(cell_el, "m:v", NS_MAP)

    if val_el is None or val_el.text is None:
        return None

    cell_type = cell_el.get("t")

    try:
        if cell_type == "s":  # Shared String
            return shared_strings[int(val_el.text)]
        if cell_type == "b":  # Boolean
            return bool(int(val_el.text))
        if cell_type == "str":  # Inline String
            return val_el.text
        if cell_type == "e":  # Error
            logger.debug(f"Ignoring cell with error value: {val_el.text}")
            return None
        if cell_type == "n" or cell_type is None:  # Number
            try:
                if "." in val_el.text:
                    return float(val_el.text)
                return int(val_el.text)
            except ValueError:
                return val_el.text
        return val_el.text
    except (IndexError, ValueError, TypeError) as e:
        logger.warning(f"Failed to parse cell value (type={cell_type}, val='{val_el.text}'): {e}")
        return None


def _parse_shared_strings(xml_root: ET.Element | None) -> list[str]:
    """Parses the sharedStrings.xml file into a simple list."""
    if xml_root is None:
        return []

    strings: list[str] = []

    for si_el in _findall_xml_elements(xml_root, "m:si", NS_MAP):
        text_parts = [t.text for t in _findall_xml_elements(si_el, ".//m:t", NS_MAP) if t.text]
        strings.append("".join(text_parts))
    return strings


def _parse_worksheet(xml_root: ET.Element | None, shared_strings: list[str]) -> dict[str, Any]:
    """Parses a single sheetN.xml file for dimensions and column headers."""
    if xml_root is None:
        return {"row_count": None, "column_count": None, "column_names": []}

    dim_el = _find_xml_element(xml_root, "m:dimension", NS_MAP)
    row_count, col_count = _parse_dimension_ref(dim_el.get("ref") if dim_el is not None else None)

    column_names: list[str] = []
    row_1 = _find_xml_element(xml_root, "m:sheetData/m:row[@r='1']", NS_MAP)

    if row_1 is not None:
        for cell_el in _findall_xml_elements(row_1, "m:c", NS_MAP):
            val = _get_cell_value(cell_el, shared_strings)
            if val is not None:
                column_names.append(str(val))

    return {"row_count": row_count, "column_count": col_count, "column_names": column_names}


def _parse_excel_properties(
    workbook_root: ET.Element | None,
    workbook_rels_root: ET.Element | None,
    shared_strings_root: ET.Element | None,
    z: zipfile.ZipFile,
) -> dict[str, Any]:
    """Parses workbook.xml and all associated worksheets. Returns a dict."""

    if workbook_root is None:
        return {"active_sheet_name": None, "sheet_names": [], "sheets": [], "has_macros": None}

    shared_strings = _parse_shared_strings(shared_strings_root)

    rId_to_path_map: dict[str, str] = {}
    if workbook_rels_root is not None:
        for el in workbook_rels_root.findall(f"{RELS_NS}Relationship"):
            if el.get("Type") == WORKSHEET_TYPE:
                rId = el.get("Id")
                target = el.get("Target")
                if rId and target:
                    rId_to_path_map[rId] = f"xl/{target.lstrip('/')}"

    sheet_names: list[str] = []
    sheet_info_list: list[dict[str, Any]] = []

    sheets_el = _find_xml_element(workbook_root, "m:sheets", NS_MAP)

    if sheets_el is not None:
        for sheet_el in _findall_xml_elements(sheets_el, "m:sheet", NS_MAP):
            name = sheet_el.get("name")
            rId = sheet_el.get(f"{R_NS}id")
            if name and rId:
                sheet_info_list.append(
                    {
                        "name": name,
                        "rId": rId,
                        "state": sheet_el.get("state"),
                    }
                )
                sheet_names.append(name)

    active_sheet_name: str | None = None
    view_el = _find_xml_element(workbook_root, "m:bookViews/m:workbookView", NS_MAP)

    if view_el is not None:
        try:
            active_tab_index = int(view_el.get("activeTab", "0"))
            if 0 <= active_tab_index < len(sheet_names):
                active_sheet_name = sheet_names[active_tab_index]
        except (ValueError, TypeError):
            pass

    sheets: list[dict[str, Any]] = []
    for sheet_info in sheet_info_list:
        sheet_path = rId_to_path_map.get(sheet_info["rId"])
        if not sheet_path:
            logger.warning(
                f"Sheet '{sheet_info['name']}' (rId: {sheet_info['rId']}) found in workbook but not in rels."
            )
            continue

        sheet_root = _read_xml_from_zip(z, sheet_path)

        sheet_data_dict = _parse_worksheet(sheet_root, shared_strings)

        sheet_data_dict["name"] = sheet_info["name"]
        sheet_data_dict["is_hidden"] = sheet_info.get("state") == "hidden"

        sheets.append(sheet_data_dict)

    return {
        "active_sheet_name": active_sheet_name,
        "sheet_names": sheet_names,
        "sheets": sheets,
        "has_macros": None,
    }


def extract_xlsx_metadata(file_path: str) -> dict[str, Any] | None:
    """
    Extracts metadata from a .xlsx file and returns it as a dictionary.
    """
    final_data: dict[str, Any] = {}
    excel_specific_data: dict[str, Any] = {}
    is_password_protected = False

    core_props_path: str | None = None
    app_props_path: str | None = None
    shared_strings_path: str | None = None

    try:
        with zipfile.ZipFile(file_path, "r") as z:
            rels_root = _read_xml_from_zip(z, "_rels/.rels")
            if rels_root is None:
                logger.warning("File is not a valid OOXML (missing _rels/.rels): '%s'", file_path)
                return None

            core_props_path = _find_relationship_target(rels_root, CORE_PROPS_TYPE)
            app_props_path = _find_relationship_target(rels_root, APP_PROPS_TYPE)

            if core_props_path is None:
                core_props_path = DEFAULT_CORE_PROPS_PATH
            if app_props_path is None:
                app_props_path = DEFAULT_APP_PROPS_PATH

            core_root = _read_xml_from_zip(z, core_props_path)
            app_root = _read_xml_from_zip(z, app_props_path)
            ct_root = _read_xml_from_zip(z, "[Content_Types].xml")
            custom_props_root = _read_xml_from_zip(z, "docProps/custom.xml")

            workbook_root = _read_xml_from_zip(z, "xl/workbook.xml")
            workbook_rels_root = _read_xml_from_zip(z, "xl/_rels/workbook.xml.rels")

            if workbook_rels_root is not None:
                ss_target = _find_relationship_target(workbook_rels_root, SHARED_STRINGS_TYPE)
                if ss_target:
                    shared_strings_path = f"xl/{ss_target.lstrip('/')}"

            shared_strings_root = _read_xml_from_zip(z, shared_strings_path or "xl/sharedStrings.xml")

            core_data = _parse_core_properties(core_root, NS_MAP)
            app_data, _ = _parse_app_properties(app_root, NS_MAP)
            ct_data = _parse_content_types_xml(ct_root)
            custom_data = _parse_custom_properties_xml(custom_props_root, NS_MAP)

            excel_specific_data = _parse_excel_properties(workbook_root, workbook_rels_root, shared_strings_root, z)

            final_data.update(core_data)
            final_data.update(app_data)
            final_data.update(ct_data)
            final_data.update(custom_data)

            if "structural_parts" in final_data:
                parts = final_data.get("structural_parts", [])
                if (
                    "application/vnd.ms-excel.sheet.macroEnabled.main+xml" in parts
                    or "application/vnd.ms-office.vbaProject" in parts
                ):
                    excel_specific_data["has_macros"] = True
                elif parts:
                    excel_specific_data["has_macros"] = False

    except zipfile.BadZipFile:
        logger.warning("File is not a valid ZIP (corrupt, password-protected, or owner file): '%s'", file_path)
        is_password_protected = True
    except Exception as e:
        logger.error(f"Unexpected error opening xlsx zip file '{file_path}': {e}", exc_info=True)
        return None

    final_data["is_password_protected"] = is_password_protected
    final_data["word"] = None
    final_data["excel"] = excel_specific_data
    final_data["powerpoint"] = None

    found_data_keys = [
        k
        for k, v in final_data.items()
        if (v not in [None, [], {}]) and (k not in ["is_password_protected", "word", "excel", "powerpoint"])
    ]
    found_excel_keys = [k for k, v in excel_specific_data.items() if (v not in [None, [], {}])]

    if not found_data_keys and not found_excel_keys:
        if is_password_protected:
            return {"is_password_protected": True}

        logger.warning("Could not extract any metadata from '%s'. File is likely minimal or empty.", file_path)
        return None

    return final_data



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
import typer
import pathlib
import datetime
import json
from typing import Annotated, Any, Optional, TYPE_CHECKING

from dorsal.common import constants

if TYPE_CHECKING:
    from dorsal.file.dorsal_file import LocalFile


logger = logging.getLogger(__name__)


def scan_file(
    ctx: typer.Context,
    path: Annotated[
        pathlib.Path,
        typer.Argument(
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="The path to the file to scan.",
        ),
    ],
    use_cache: Annotated[
        bool,
        typer.Option(
            "--use-cache",
            help="Force the use of the cache, overriding any global setting.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    skip_cache: Annotated[
        bool,
        typer.Option(
            "--skip-cache",
            help="Bypass the local cache and re-process the file.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    overwrite_cache: Annotated[
        bool,
        typer.Option(
            "--overwrite-cache",
            help="Re-process the file and overwrite the local cache with new data.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    json_output: Annotated[
        bool,
        typer.Option(
            "--json",
            help="Output JSON to stdout. Can be combined with --save.",
            rich_help_panel="Output Options",
        ),
    ] = False,
    save: Annotated[
        bool,
        typer.Option(
            "-s",
            "--save",
            help="Save a JSON report to the default directory or --output path.",
            rich_help_panel="Output Options",
        ),
    ] = False,
    report: Annotated[
        bool,
        typer.Option(
            "--report",
            help="Generate a self-contained HTML report to the default directory or --output path.",
            rich_help_panel="Output Options",
        ),
    ] = False,
    output_path: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            "-o",
            "--output",
            help="Custom output path (file or directory) for generated reports.",
            dir_okay=True,
            file_okay=True,
            writable=True,
            resolve_path=True,
            rich_help_panel="Output Options",
        ),
    ] = None,
    template: Annotated[
        str,
        typer.Option(
            "--template",
            "-t",
            help="Name or path of the report template to use.",
            rich_help_panel="Output Options",
        ),
    ] = "default",
    resolve_links: Annotated[
        bool,
        typer.Option(
            "--follow-links/--no-follow-links",
            help="Follow symlinks to scan target content vs scanning the link itself.",
        ),
    ] = True,
):
    """
    Scans and displays the full metadata for a file, with options to save reports.
    """
    from dorsal.cli.views.file import create_file_info_panel
    from dorsal.common.cli import (
        exit_cli,
        EXIT_CODE_ERROR,
        get_rich_console,
        determine_use_cache_value,
    )
    from dorsal.file.dorsal_file import LocalFile

    if use_cache and skip_cache:
        exit_cli(
            code=EXIT_CODE_ERROR,
            message="Error: --use-cache and --skip-cache cannot be used together.",
        )

    if skip_cache and overwrite_cache:
        exit_cli(
            code=EXIT_CODE_ERROR,
            message="Error: --skip-cache and --overwrite-cache cannot be used together.",
        )

    if json_output and report:
        exit_cli(
            code=EXIT_CODE_ERROR,
            message="Error: --json (stdout) and --report (HTML) flags are not compatible.",
        )

    console = get_rich_console()
    palette = ctx.obj.get("palette", {})

    if output_path and not (save or report):
        if str(output_path).lower().endswith(".json"):
            save = True
        if str(output_path).lower().endswith(".html"):
            report = True
        if not (save or report):
            console.print(
                f"⚠️ [yellow]Warning:[/] --output path '{output_path}' was specified, but no report"
                f" type was requested. (Did you forget --save or --report?)",
                style=palette.get("warning", "yellow"),
            )

    use_cache_value = determine_use_cache_value(use_cache=use_cache, skip_cache=skip_cache)
    palette = ctx.obj["palette"]

    if not json_output:
        console.print(f"📄 Scanning metadata for [{palette['primary_value']}]{path.name}[/]")

    try:
        local_file = LocalFile(
            file_path=str(path),
            use_cache=use_cache_value,
            overwrite_cache=overwrite_cache,
            follow_symlinks=resolve_links,
        )

        record_dict: dict[str, Any] = local_file.to_dict(mode="json")
        if "local_attributes" in record_dict:
            record_dict["local_filesystem"] = record_dict["local_attributes"]
            record_dict["local_filesystem"]["full_path"] = record_dict["local_attributes"].get("file_path", str(path))

            for key in ["date_created", "date_modified", "date_accessed"]:
                val = record_dict["local_filesystem"].get(key)
                if isinstance(val, datetime.datetime):
                    record_dict["local_filesystem"][key] = val.isoformat()

        else:
            record_dict["local_filesystem"] = {
                "full_path": local_file._file_path,
                "date_created": (local_file.date_created.isoformat() if hasattr(local_file, "date_created") else None),
                "date_modified": (
                    local_file.date_modified.isoformat() if hasattr(local_file, "date_modified") else None
                ),
            }

        if json_output:
            console.print(json.dumps(record_dict, indent=2, default=str, ensure_ascii=False))

        else:
            panel = create_file_info_panel(
                record_dict=record_dict,
                title=f"File Record: {local_file.name}",
                palette=palette,
                private=None,
                source=local_file._source,
            )
            console.print(panel)

        if save:
            _save_json_report(
                ctx=ctx,
                record=record_dict,
                original_path=path,
                output_path=output_path,
                palette=palette,
            )

        if report:
            _save_html_report(
                local_file=local_file,
                original_path=path,
                output_path=output_path,
                palette=palette,
                template=template,
            )

    except typer.Exit:
        raise
    except Exception as err:
        logger.exception(f"CLI 'scan' command failed while processing {path}.")
        exit_cli(code=EXIT_CODE_ERROR, message=f"An unexpected error occurred: {err}")


def _get_final_path(original_path: pathlib.Path, output_path: Optional[pathlib.Path], suffix: str) -> pathlib.Path:
    """Helper to determine the final save path for a report."""

    if output_path:
        if output_path.is_dir():
            return output_path / f"{original_path.stem}_report{suffix}"
        else:
            return output_path

    constants.CLI_SCAN_REPORTS_DIR.mkdir(parents=True, exist_ok=True)
    timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    safe_filename = original_path.name.replace(" ", "_")
    return constants.CLI_SCAN_REPORTS_DIR / f"{safe_filename}-{timestamp}{suffix}"


def _save_json_report(
    ctx: typer.Context,
    record: dict,
    original_path: pathlib.Path,
    output_path: Optional[pathlib.Path],
    palette: dict,
):
    """Saves the JSON scan report to a specified or default path."""
    from dorsal.common.cli import get_rich_console

    console = get_rich_console()

    num_reports = ctx.params.get("save", 0) + ctx.params.get("report", 0)
    if output_path and not output_path.is_dir() and num_reports > 1:
        console.print(
            "⚠️ [yellow]Warning:[/] Both --save and --report are specified, but --output"
            " is a single file. Reports may overwrite each other.",
            style=palette.get("warning", "yellow"),
        )

    final_path = _get_final_path(original_path, output_path, ".json")

    try:
        final_path.parent.mkdir(parents=True, exist_ok=True)
        with open(final_path, "w", encoding="utf-8") as f:
            json.dump(record, f, indent=2, default=str, ensure_ascii=False)

        if not ctx.params.get("json_output"):
            console.print(f"✅ JSON report saved to: [{palette.get('primary_value')}]{final_path}[/]")
    except Exception as e:
        logger.error(f"Failed to save JSON report: {e}")
        console.print(
            f"⚠️ Could not save JSON report to {final_path}. Error: {e}", style=palette.get("warning", "yellow")
        )


def _save_html_report(
    local_file: LocalFile,
    original_path: pathlib.Path,
    output_path: Optional[pathlib.Path],
    palette: dict,
    template: str,
):
    """Handles the logic for generating and saving the HTML report."""
    from dorsal.cli.views.file import create_file_info_panel
    from dorsal.common.cli import get_rich_console
    from dorsal.file.dorsal_file import LocalFile
    from dorsal.api.file import generate_html_file_report

    console = get_rich_console()
    final_path = _get_final_path(original_path, output_path, ".html")

    try:
        with console.status(f"📄 Generating HTML report for '[bold]{original_path.name}[/]'..."):
            generate_html_file_report(
                file_path=local_file._file_path,
                local_file=local_file,
                output_path=str(final_path),
                template=template,
            )
        console.print(f"✅ HTML report saved to: [{palette.get('primary_value')}]{final_path}[/]")

    except Exception as e:
        logger.error(f"Failed to generate HTML report: {e}")
        console.print(f"⚠️ Could not generate HTML report. Error: {e}", style=palette.get("warning", "yellow"))



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from unittest.mock import patch, MagicMock
import typer

from dorsal.common import cli


@patch("typer.secho")
def test_exit_cli_success_with_message(mock_secho):
    """Test exiting with a success code and a message."""
    with pytest.raises(typer.Exit) as excinfo:
        cli.exit_cli(code=0, message="Operation successful.")

    assert excinfo.value.exit_code == 0
    mock_secho.assert_called_once_with("Operation successful.", err=True)


@patch("typer.secho")
def test_exit_cli_error_with_message(mock_secho):
    """Test exiting with an error code and ensuring the message is prefixed."""
    with pytest.raises(typer.Exit) as excinfo:
        cli.exit_cli(code=1, message="File not found.")

    assert excinfo.value.exit_code == 1
    mock_secho.assert_called_once_with("Error: File not found.", fg=typer.colors.RED, err=True)


@patch("typer.secho")
def test_exit_cli_no_message(mock_secho):
    """Test that no message is printed if none is provided."""
    with pytest.raises(typer.Exit) as excinfo:
        cli.exit_cli(code=5)

    assert excinfo.value.exit_code == 5
    mock_secho.assert_not_called()


@pytest.mark.parametrize(
    "use_cache_flag, skip_cache_flag, expected_arg",
    [
        (True, False, True),
        (False, True, False),
        (True, True, True),
        (False, False, None),
    ],
)
@patch("dorsal.file.cache.config.get_cache_enabled")
def test_determine_use_cache_value(mock_get_cache_enabled, use_cache_flag, skip_cache_flag, expected_arg):
    """Test the logic for resolving cache flags."""
    mock_get_cache_enabled.side_effect = lambda use_cache: use_cache

    cli.determine_use_cache_value(use_cache=use_cache_flag, skip_cache=skip_cache_flag)

    mock_get_cache_enabled.assert_called_once_with(use_cache=expected_arg)


def test_parse_cli_options_empty():
    """Test that empty or None options return an empty dictionary."""
    assert cli.parse_cli_options(None, {}) == {}
    assert cli.parse_cli_options([], {}) == {}


@patch("dorsal.common.cli.get_error_console")
def test_parse_cli_options_malformed(mock_get_error_console):
    """Test that options missing an '=' sign are skipped and a warning is printed."""
    mock_console = MagicMock()
    mock_get_error_console.return_value = mock_console

    options = ["valid=true", "invalid_no_equals"]
    palette = {"panel_title_warning": "yellow"}

    result = cli.parse_cli_options(options, palette)

    assert result == {"valid": True}
    mock_console.print.assert_called_once()
    assert "Skipping malformed option 'invalid_no_equals'" in mock_console.print.call_args[0][0]


def test_parse_cli_options_basic_types():
    """Test parsing of standard types: booleans, nulls, ints, floats, and strings."""
    options = [
        "is_true=true",
        "is_yes=yes",
        "is_false=false",
        "is_no=no",
        "is_null=null",
        "is_none=none",
        "integer=42",
        "floating=3.14",
        "string=hello world",
        "mixed_case_bool=TrUe",
    ]

    expected = {
        "is_true": True,
        "is_yes": True,
        "is_false": False,
        "is_no": False,
        "is_null": None,
        "is_none": None,
        "integer": 42,
        "floating": 3.14,
        "string": "hello world",
        "mixed_case_bool": True,
    }

    assert cli.parse_cli_options(options, {}) == expected


def test_parse_cli_options_json_structures():
    """Test parsing of JSON dictionaries and arrays, including fallbacks for invalid JSON."""
    options = [
        'vad_filter={"threshold": 0.8, "min_speech_duration_ms": 250}',
        'array_val=[1, 2, "three"]',
        'invalid_json={"missing_bracket": true',
        "not_quite_json={just some text}",
    ]

    expected = {
        "vad_filter": {"threshold": 0.8, "min_speech_duration_ms": 250},
        "array_val": [1, 2, "three"],
        "invalid_json": '{"missing_bracket": true',
        "not_quite_json": "{just some text}",
    }

    assert cli.parse_cli_options(options, {}) == expected



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import pytest
from unittest.mock import MagicMock, patch
from dorsal.file.dorsal_file import LocalFile
from dorsal.common.exceptions import AttributeConflictError, FileAnnotatorError
from pydantic import ValidationError
from types import SimpleNamespace


@pytest.fixture
def mock_local_file():
    """Creates a LocalFile with a mocked internal model."""
    with patch("dorsal.file.dorsal_file.LocalFile.__init__", return_value=None):
        lf = LocalFile("dummy.txt")

        lf._file_path = "dummy.txt"
        lf.hash = "a" * 64
        lf.validation_hash = "b" * 64
        lf._model_runner = MagicMock()

        lf.model = MagicMock()
        lf.model.annotations = SimpleNamespace()

        lf.model.annotations.__pydantic_extra__ = {}
        lf.model.annotations.model_fields_set = set()

        lf.pdf = None
        lf.mediainfo = None
        lf.ebook = None
        lf.office = None

        lf._populate = MagicMock()

        return lf


@patch("dorsal.file.file_annotator.FILE_ANNOTATOR")
def test_annotate_using_pipeline_step_success(mock_annotator, mock_local_file):
    """Tests successful annotation via pipeline step config."""
    step_config = {"annotation_model": ("mod", "cls"), "schema_id": "test/schema"}
    mock_annotation = MagicMock()

    mock_annotation.source.id = "pipeline_source"
    mock_annotation.source.version = "1.0"
    mock_annotation.source.variant = None

    mock_annotator.annotate_file_using_pipeline_step.return_value = mock_annotation

    mock_local_file._annotate_using_pipeline_step(pipeline_step_config=step_config)

    assert getattr(mock_local_file.model.annotations, "test/schema") == [mock_annotation]


def test_annotate_using_pipeline_step_invalid_schema(mock_local_file):
    """Tests error with invalid schema ID."""

    with pytest.raises(ValidationError):
        mock_local_file._annotate_using_pipeline_step(
            pipeline_step_config={"schema_id": "bad_id", "annotation_model": ("a", "b")}
        )


@patch("dorsal.file.file_annotator.FILE_ANNOTATOR")
def test_annotate_model_validator_success(mock_annotator, mock_local_file):
    """Tests successful annotation via explicit model/validator."""
    mock_annotation = MagicMock()
    mock_annotation.source.id = "manual_source"
    mock_annotation.source.version = "1.0"
    mock_annotation.source.variant = None

    mock_annotator.annotate_file_using_model_and_validator.return_value = mock_annotation

    dummy_model_cls = MagicMock()
    dummy_model_cls.__name__ = "DummyModel"
    dummy_model_cls.__str__ = lambda x: "DummyModel"

    dummy_validator = MagicMock()
    dummy_validator.__name__ = "DummyValidator"
    dummy_validator.__str__ = lambda x: "DummyValidator"

    mock_local_file._annotate_using_model_and_validator(
        schema_id="test/manual",
        annotation_model=dummy_model_cls,
        validation_model=dummy_validator,
        overwrite=True,
    )

    mock_annotator.annotate_file_using_model_and_validator.assert_called_once()

    assert getattr(mock_local_file.model.annotations, "test/manual") == [mock_annotation]


@patch("dorsal.file.file_annotator.FILE_ANNOTATOR")
def test_annotate_model_validator_failure(mock_annotator, mock_local_file):
    """Tests that FileAnnotatorError is re-raised."""
    mock_annotator.annotate_file_using_model_and_validator.side_effect = FileAnnotatorError("Boom")

    dummy_model = MagicMock()
    dummy_model.__name__ = "DummyModel"

    with pytest.raises(FileAnnotatorError):
        mock_local_file._annotate_using_model_and_validator(schema_id="test/fail", annotation_model=dummy_model)


def test_remove_annotation_missing(mock_local_file):
    """Tests removing an annotation that doesn't exist (safe no-op)."""
    mock_local_file.remove_annotation("non/existent")
    mock_local_file._populate.assert_not_called()


def test_remove_annotation_success(mock_local_file):
    """Tests successful removal."""

    mock_ann = MagicMock()
    mock_ann.source.id = "delete_me"

    setattr(mock_local_file.model.annotations, "test/remove", [mock_ann])

    mock_local_file.remove_annotation("test/remove")

    assert not hasattr(mock_local_file.model.annotations, "test/remove")
    mock_local_file._populate.assert_called_once()


def test_set_annotation_conflict(mock_local_file):
    """Tests that overwriting without overwrite=True raises error."""
    schema_id = "test/conflict"

    existing_mock = MagicMock()
    existing_mock.source.id = "conflict_id"
    existing_mock.source.version = "1.0"
    existing_mock.source.variant = None

    setattr(mock_local_file.model.annotations, schema_id, [existing_mock])

    new_mock = MagicMock()
    new_mock.source.id = "conflict_id"
    new_mock.source.version = "1.0"
    new_mock.source.variant = None

    with pytest.raises(AttributeConflictError):
        mock_local_file._set_annotation_attribute(schema_id=schema_id, annotation=new_mock, overwrite=False)


def test_get_annotations_returns_list(mock_local_file):
    """Test the standard list retrieval behavior."""

    schema_id = "test/multi"
    ann1 = MagicMock()
    ann1.source.id = "source_1"
    ann2 = MagicMock()
    ann2.source.id = "source_2"

    setattr(mock_local_file.model.annotations, schema_id, [ann1, ann2])

    all_anns = mock_local_file.get_annotations(schema_id)
    assert isinstance(all_anns, list)
    assert len(all_anns) == 2

    filtered = mock_local_file.get_annotations(schema_id, source_id="source_1")
    assert len(filtered) == 1
    assert filtered[0].source.id == "source_1"

    empty = mock_local_file.get_annotations("test/missing")
    assert isinstance(empty, list)
    assert len(empty) == 0


def test_get_latest_annotation_success(mock_local_file):
    """Test retrieving the latest annotation by date_modified."""
    schema_id = "test/dates"
    now = datetime.datetime.now(datetime.UTC)

    ann_old = MagicMock()
    ann_old.date_modified = now - datetime.timedelta(days=1)
    ann_old.record.data = "old"

    ann_new = MagicMock()
    ann_new.date_modified = now
    ann_new.record.data = "new"

    ann_mid = MagicMock()
    ann_mid.date_modified = now - datetime.timedelta(hours=1)
    ann_mid.record.data = "mid"

    setattr(mock_local_file.model.annotations, schema_id, [ann_old, ann_new, ann_mid])

    latest = mock_local_file.get_latest_annotation(schema_id)

    assert latest is not None
    assert latest.record.data == "new"


def test_get_latest_annotation_with_source_filter(mock_local_file):
    """Test retrieving latest annotation filtered by source ID first."""
    schema_id = "test/dates_source"
    now = datetime.datetime.now(datetime.UTC)

    ann_wrong_source = MagicMock()
    ann_wrong_source.date_modified = now + datetime.timedelta(hours=1)
    ann_wrong_source.source.id = "src_B"

    ann_src_a_new = MagicMock()
    ann_src_a_new.date_modified = now
    ann_src_a_new.source.id = "src_A"

    ann_src_a_old = MagicMock()
    ann_src_a_old.date_modified = now - datetime.timedelta(days=1)
    ann_src_a_old.source.id = "src_A"

    setattr(mock_local_file.model.annotations, schema_id, [ann_wrong_source, ann_src_a_new, ann_src_a_old])

    latest = mock_local_file.get_latest_annotation(schema_id, source_id="src_A")

    assert latest is not None
    assert latest.source.id == "src_A"
    assert latest == ann_src_a_new


def test_get_latest_annotation_empty(mock_local_file):
    """Test returns None when no annotations exist."""
    assert mock_local_file.get_latest_annotation("test/empty") is None


def test_get_latest_annotation_fallback_no_dates(mock_local_file):
    """Test fallback behavior when date_modified is missing (e.g. freshly created local objects)."""
    schema_id = "test/nodates"

    ann1 = MagicMock()
    del ann1.date_modified
    ann1.record.id = 1

    ann2 = MagicMock()
    del ann2.date_modified
    ann2.record.id = 2

    setattr(mock_local_file.model.annotations, schema_id, [ann1, ann2])

    result = mock_local_file.get_latest_annotation(schema_id)
    assert result is not None

    assert result in [ann1, ann2]



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from unittest.mock import patch, MagicMock
from importlib.metadata import PackageNotFoundError

from dorsal.registry import resolution
from dorsal.common.exceptions import DorsalError, AuthError


@pytest.fixture
def mock_dorsal_client():
    """Mock the registry client to avoid network calls."""
    with patch("dorsal.registry.resolution.get_shared_dorsal_client") as mock:
        yield mock.return_value


@pytest.fixture
def mock_find_package():
    """Mock the local config lookup."""
    with patch("dorsal.registry.resolution.find_package_name_by_class") as mock:
        yield mock


def test_resolve_registry_id_success(mock_dorsal_client):
    """
    If the target looks like a Registry ID (user/repo), we should fetch it.
    """

    target = "dorsalhub/whisper"
    mock_dorsal_client.get_registry_model.return_value.package_name = "dorsal-whisper-pkg"

    strategy, pkg_name = resolution.resolve_target(target)

    mock_dorsal_client.get_registry_model.assert_called_with(target)
    assert strategy == "registry_id"
    assert pkg_name == "dorsal-whisper-pkg"


def test_resolve_registry_id_auth_error(mock_dorsal_client):
    """
    If the registry returns an AuthError (401/403), it should bubble up
    so the CLI can prompt the user to login.
    """
    target = "private/secret-model"
    mock_dorsal_client.get_registry_model.side_effect = AuthError("Unauthorized")

    with pytest.raises(AuthError):
        resolution.resolve_target(target)


def test_resolve_registry_id_general_error(mock_dorsal_client):
    """
    General network/api errors should be wrapped in DorsalError.
    """
    target = "broken/model"
    mock_dorsal_client.get_registry_model.side_effect = Exception("500 Server Error")

    with pytest.raises(DorsalError) as exc:
        resolution.resolve_target(target)

    assert "Failed to resolve model" in str(exc.value)


def test_resolve_class_name_from_config(mock_find_package):
    """
    If the target is NOT a registry ID, checks if it maps to a known class in config.
    """

    target = "MyCoolModel"
    mock_find_package.return_value = "dorsal-my-cool-model"

    strategy, pkg_name = resolution.resolve_target(target)

    mock_find_package.assert_called_with(target)

    assert strategy == "package"
    assert pkg_name == "dorsal-my-cool-model"


def test_resolve_fallthrough_to_class_name(mock_find_package):
    """
    If it's not a registry ID and not in the config, assume it's a raw class name/package reference.
    """
    target = "UnknownIdentifier"
    mock_find_package.return_value = None

    strategy, pkg_name = resolution.resolve_target(target)

    assert strategy == "class_name"



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import copy
import json
from typing import Any, Protocol, TYPE_CHECKING, runtime_checkable

from dorsal.common.constants import ANNOTATION_MAX_SIZE_BYTES
from dorsal.common.exceptions import PydanticValidationError
from dorsal.file.utils.size import human_filesize, check_record_size


if TYPE_CHECKING:
    from dorsal.file.validators.file_record import AnnotationGroup, ShardedAnnotation

logger = logging.getLogger(__name__)


@runtime_checkable
class ShardingStrategy(Protocol):
    fields_to_drop: list[str]

    def split(self, record: dict[str, Any]) -> list[dict[str, Any]]: ...


class ListBasedStrategy:
    """
    Splits a record by distributing a high-cardinality list field across multiple chunks.
    Optimized for O(N) performance using running byte counts.
    """

    def __init__(self, list_field: str, fields_to_drop_in_successors: list[str] | None = None):
        self.list_field = list_field
        self.fields_to_drop = fields_to_drop_in_successors or []

    def split(self, record: dict[str, Any]) -> list[dict[str, Any]]:
        items = record.get(self.list_field)

        if not isinstance(items, list) or not items:
            return [record]

        template_0 = copy.deepcopy(record)
        template_0[self.list_field] = []

        template_n = copy.deepcopy(template_0)
        for field in self.fields_to_drop:
            if field in template_n:
                del template_n[field]

        size_template_0 = check_record_size(template_0)
        size_template_n = check_record_size(template_n)

        if size_template_0 > ANNOTATION_MAX_SIZE_BYTES:
            raise ValueError(f"Metadata header is too large ({size_template_0} bytes) to allow sharding.")

        chunks: list[dict[str, Any]] = []

        current_items: list[Any] = []
        current_chunk_size = size_template_0
        is_first_chunk = True

        for item in items:
            item_size = check_record_size(item)
            addition_cost = item_size if not current_items else (item_size + 1)

            if (current_chunk_size + addition_cost) > ANNOTATION_MAX_SIZE_BYTES:
                if not current_items:
                    raise ValueError(
                        f"A single item in '{self.list_field}' exceeds the {human_filesize(ANNOTATION_MAX_SIZE_BYTES)} limit. Cannot shard."
                    )

                base = template_0 if is_first_chunk else template_n
                chunk = base.copy()

                chunk[self.list_field] = current_items
                chunks.append(chunk)

                is_first_chunk = False
                current_items = []
                current_chunk_size = size_template_n
                addition_cost = item_size

                if (current_chunk_size + addition_cost) > ANNOTATION_MAX_SIZE_BYTES:
                    raise ValueError(
                        f"A single item in '{self.list_field}' exceeds the {human_filesize(ANNOTATION_MAX_SIZE_BYTES)} limit. Cannot shard."
                    )

            current_items.append(item)
            current_chunk_size += addition_cost

        if current_items:
            base = template_0 if is_first_chunk else template_n
            chunk = base.copy()
            chunk[self.list_field] = current_items
            chunks.append(chunk)

        return chunks


class StringShardingStrategy:
    """
    Splits a record by chunking a massive string field (e.g. LLM response text).
    Handles UTF-8 boundaries and JSON escaping overhead.
    """

    def __init__(self, text_field: str, fields_to_drop_in_successors: list[str] | None = None):
        self.text_field = text_field
        self.fields_to_drop = fields_to_drop_in_successors or []

    def split(self, record: dict[str, Any]) -> list[dict[str, Any]]:
        text = record.get(self.text_field, "")
        if not text or not isinstance(text, str):
            return [record]

        base_template = copy.deepcopy(record)
        base_template[self.text_field] = ""

        skeleton_size = check_record_size(base_template)

        safety_buffer = 4096

        capacity = ANNOTATION_MAX_SIZE_BYTES - skeleton_size - safety_buffer

        if capacity < 1024:
            raise ValueError(
                f"Metadata header is too large ({skeleton_size} bytes). "
                "Not enough room left to shard text content safely."
            )

        chunks: list[dict[str, Any]] = []

        text_bytes = text.encode("utf-8")
        total_bytes = len(text_bytes)
        offset = 0

        while offset < total_bytes:
            end = min(offset + capacity, total_bytes)
            while end > offset and end < total_bytes and (text_bytes[end] & 0xC0) == 0x80:
                end -= 1

            if end == offset:
                raise ValueError("Unable to split text: single character exceeds chunk capacity.")

            chunk_str = text_bytes[offset:end].decode("utf-8")
            chunk_record = copy.deepcopy(base_template)
            chunk_record[self.text_field] = chunk_str

            if offset > 0:
                for f in self.fields_to_drop:
                    chunk_record.pop(f, None)

            if check_record_size(chunk_record) > ANNOTATION_MAX_SIZE_BYTES:
                logger.debug("Chunk exceeded limit due to escaping overhead. Shrinking and retrying.")
                capacity = int(capacity * 0.9)
                if capacity < 100:
                    raise ValueError("Cannot shrink chunk further; text contains extremely high-overhead characters.")
                continue

            chunks.append(chunk_record)
            offset = end

        return chunks


SHARDING_REGISTRY: dict[str, ShardingStrategy] = {
    "open/document-extraction": ListBasedStrategy(list_field="blocks"),
    "open/object-detection": ListBasedStrategy(list_field="objects"),
    "open/entity-extraction": ListBasedStrategy(list_field="entities"),
    "open/classification": ListBasedStrategy(list_field="labels"),
    "open/audio-transcription": ListBasedStrategy(list_field="segments", fields_to_drop_in_successors=["text"]),
    "open/regression": ListBasedStrategy(list_field="points"),
    "open/llm-output": StringShardingStrategy(
        text_field="response_data", fields_to_drop_in_successors=["prompt", "generation_params"]
    ),
    # Unsupported
    # "open/geolocation" - no appropriate sharding strategy yet
    # "open/generic" - never breaches 1MiB
    # "open/embedding" - never breaches 1MiB
}


def process_record_for_sharding(schema_id: str | None, record: dict[str, Any]) -> list[dict[str, Any]]:
    """
    Checks if a record exceeds the global size limit and shards it if a strategy exists.
    """
    size = check_record_size(record)

    if size <= ANNOTATION_MAX_SIZE_BYTES:
        return [record]

    if schema_id is None:
        raise ValueError(
            f"Record exceeds limit ({human_filesize(size)}) but no schema_id was provided to determine a sharding strategy."
        )

    strategy = SHARDING_REGISTRY.get(schema_id)

    if strategy is None:
        raise ValueError(
            f"Record for '{schema_id}' exceeds limit ({human_filesize(size)}) and this schema does not support sharding. Please reduce the record size."
        )

    logger.info(
        "Record for '%s' exceeds limit (%s). Attempting sharding strategy: %s...",
        schema_id,
        human_filesize(size),
        strategy.__class__.__name__,
    )

    chunks = strategy.split(record)

    logger.info("Successfully sharded record into %d chunks.", len(chunks))
    return chunks


def reassemble_record(group: AnnotationGroup) -> tuple[str, dict[str, Any]]:
    """
    Reassembles a sharded AnnotationGroup into a single atomic record.
    Returns: (schema_id, record_dict)
    """
    from dorsal.file.validators.file_record import ShardedAnnotation

    if not group.annotations:
        raise ValueError("Cannot reassemble an empty AnnotationGroup.")

    try:
        chunks = [ShardedAnnotation.model_validate(ann, from_attributes=True) for ann in group.annotations]
    except PydanticValidationError as e:
        raise ValueError(f"Sharding integrity error: Chunk missing required fields. Details: {e}") from e

    sorted_chunks = sorted(chunks, key=lambda x: x.group.index)
    head = sorted_chunks[0]
    total = head.group.total

    if len(sorted_chunks) != total:
        raise ValueError(f"Incomplete group: Expected {total} chunks, got {len(sorted_chunks)}")

    head_record = head.record.model_dump()
    strategy: ShardingStrategy | None = None
    matched_schema_id: str | None = None

    for schema_id, strat in SHARDING_REGISTRY.items():
        if isinstance(strat, ListBasedStrategy):
            if strat.list_field in head_record and isinstance(head_record[strat.list_field], list):
                strategy = strat
                matched_schema_id = schema_id
                break
        elif isinstance(strat, StringShardingStrategy):
            if strat.text_field in head_record and isinstance(head_record[strat.text_field], str):
                strategy = strat
                matched_schema_id = schema_id
                break

    if not strategy or matched_schema_id is None:
        raise ValueError("Could not detect sharding strategy from record structure.")

    reassembled_record = copy.deepcopy(head_record)

    if isinstance(strategy, ListBasedStrategy):
        master_list = reassembled_record[strategy.list_field]
        for chunk in sorted_chunks[1:]:
            chunk_data = chunk.record.model_dump()
            items = chunk_data.get(strategy.list_field, [])
            master_list.extend(items)

    elif isinstance(strategy, StringShardingStrategy):
        text_parts = [reassembled_record[strategy.text_field]]
        for chunk in sorted_chunks[1:]:
            chunk_data = chunk.record.model_dump()
            part = chunk_data.get(strategy.text_field, "")
            text_parts.append(part)

        reassembled_record[strategy.text_field] = "".join(text_parts)

    return matched_schema_id, reassembled_record



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import re
from unittest.mock import MagicMock


from dorsal.file.configs import model_runner
from dorsal.file.configs.model_runner import (
    check_media_type_dependency,
    check_extension_dependency,
    check_size_dependency,
    check_name_dependency,
    MediaTypeDependencyConfig,
    FileExtensionDependencyConfig,
    FileSizeDependencyConfig,
    FilenameDependencyConfig,
)


@pytest.fixture
def mock_results():
    """Factory to create a list containing a single mock RunModelResult."""

    def _create(record_data):
        mock_res = MagicMock()

        mock_res.record = record_data
        return [mock_res]

    return _create


class TestCheckMediaTypeDependency:
    """Tests for check_media_type_dependency including all unhappy paths."""

    def test_missing_base_record(self):
        """Unhappy Path: model_results[0].record is None."""
        results = [MagicMock(record=None)]
        config = MediaTypeDependencyConfig(include={"text/plain"})

        assert check_media_type_dependency(results, config) is False

    def test_record_missing_media_type_field(self, mock_results):
        """Unhappy Path: record exists but has no 'media_type' key."""
        results = mock_results({"size": 100})
        config = MediaTypeDependencyConfig(include={"text/plain"})

        assert check_media_type_dependency(results, config) is False

    def test_explicit_exclude_full_match(self, mock_results):
        """Unhappy Path: Media type matches an entry in 'exclude' list fully."""
        results = mock_results({"media_type": "application/pdf"})
        config = MediaTypeDependencyConfig(exclude={"application/pdf"})

        assert check_media_type_dependency(results, config) is False

    def test_explicit_exclude_head_match(self, mock_results):
        """Unhappy Path: Media type head (prefix) matches 'exclude'."""
        results = mock_results({"media_type": "image/png"})
        config = MediaTypeDependencyConfig(exclude={"image"})

        assert check_media_type_dependency(results, config) is False

    def test_include_rule_mismatch(self, mock_results):
        """Unhappy Path: Inclusion rules exist, but media type doesn't match."""
        results = mock_results({"media_type": "text/csv"})
        config = MediaTypeDependencyConfig(include={"application/pdf"})

        assert check_media_type_dependency(results, config) is False

    def test_pattern_rule_mismatch(self, mock_results):
        """Unhappy Path: Regex pattern provided, but doesn't match."""
        results = mock_results({"media_type": "audio/mpeg"})
        config = MediaTypeDependencyConfig(pattern=r"^image/.*")

        assert check_media_type_dependency(results, config) is False

    def test_include_full_match_success(self, mock_results):
        """Happy Path: Exact match in 'include'."""
        results = mock_results({"media_type": "application/json"})
        config = MediaTypeDependencyConfig(include={"application/json", "text/plain"})

        assert check_media_type_dependency(results, config) is True

    def test_include_head_match_success(self, mock_results):
        """Happy Path: Head match in 'include'."""
        results = mock_results({"media_type": "video/mp4"})
        config = MediaTypeDependencyConfig(include={"video"})

        assert check_media_type_dependency(results, config) is True

    def test_pattern_string_match_success(self, mock_results):
        """Happy Path: Regex string matches."""
        results = mock_results({"media_type": "text/x-python"})
        config = MediaTypeDependencyConfig(pattern=r"text/.*")

        assert check_media_type_dependency(results, config) is True

    def test_pattern_compiled_match_success(self, mock_results):
        """Happy Path: Compiled regex matches."""
        results = mock_results({"media_type": "application/vnd.ms-excel"})
        config = MediaTypeDependencyConfig(pattern=re.compile(r".*excel$"))

        assert check_media_type_dependency(results, config) is True

    def test_no_rules_passthrough(self, mock_results):
        """Happy Path: No include/exclude/pattern rules provided -> Pass."""
        results = mock_results({"media_type": "any/thing"})
        config = MediaTypeDependencyConfig()

        assert check_media_type_dependency(results, config) is True


class TestCheckExtensionDependency:
    def test_extension_match_success(self, mock_results):
        results = mock_results({"extension": ".pdf"})
        config = FileExtensionDependencyConfig(extensions={".pdf", ".txt"})

        assert check_extension_dependency(results, config) is True

    def test_extension_case_insensitivity(self, mock_results):
        """Verify extension matching handles case differences (logic usually lowercases)."""
        results = mock_results({"extension": ".PDF"})
        config = FileExtensionDependencyConfig(extensions={".pdf"})

        assert check_extension_dependency(results, config) is True

    def test_extension_mismatch(self, mock_results):
        results = mock_results({"extension": ".jpg"})
        config = FileExtensionDependencyConfig(extensions={".png"})

        assert check_extension_dependency(results, config) is False

    def test_missing_record_or_extension(self, mock_results):
        assert (
            check_extension_dependency([MagicMock(record=None)], FileExtensionDependencyConfig(extensions={".pdf"}))
            is False
        )

        assert (
            check_extension_dependency(
                mock_results({"other": "field"}), FileExtensionDependencyConfig(extensions={".pdf"})
            )
            is False
        )

        assert (
            check_extension_dependency(
                mock_results({"extension": None}), FileExtensionDependencyConfig(extensions={".pdf"})
            )
            is False
        )


class TestCheckSizeDependency:
    def test_size_within_range(self, mock_results):
        results = mock_results({"size": 500})
        config = FileSizeDependencyConfig(min_size=100, max_size=1000)

        assert check_size_dependency(results, config) is True

    def test_size_below_min(self, mock_results):
        results = mock_results({"size": 50})
        config = FileSizeDependencyConfig(min_size=100)

        assert check_size_dependency(results, config) is False

    def test_size_above_max(self, mock_results):
        results = mock_results({"size": 2000})
        config = FileSizeDependencyConfig(max_size=1000)

        assert check_size_dependency(results, config) is False

    def test_invalid_size_field(self, mock_results):
        assert check_size_dependency(mock_results({}), FileSizeDependencyConfig(min_size=1)) is False

        assert check_size_dependency(mock_results({"size": "big"}), FileSizeDependencyConfig(min_size=1)) is False


class TestCheckNameDependency:
    def test_name_match_regex_string(self, mock_results):
        results = mock_results({"name": "report_2025.pdf"})
        config = FilenameDependencyConfig(pattern=r"^report_\d+")

        assert check_name_dependency(results, config) is True

    def test_name_match_compiled_regex(self, mock_results):
        results = mock_results({"name": "test_file.py"})
        config = FilenameDependencyConfig(pattern=re.compile(r"\.py$"))

        assert check_name_dependency(results, config) is True

    def test_name_mismatch(self, mock_results):
        results = mock_results({"name": "image.png"})
        config = FilenameDependencyConfig(pattern=r"\.txt$")

        assert check_name_dependency(results, config) is False

    def test_missing_name_field(self, mock_results):
        results = mock_results({"size": 100})
        config = FilenameDependencyConfig(pattern=r".*")

        assert check_name_dependency(results, config) is False
